import numpy as np
import pandas as pd

from scipy.special import rel_entr
from sklearn.metrics import confusion_matrix, roc_auc_score
from statsmodels.tools import add_constant

from .distributions import (glm_likelihood_bernoulli, glm_likelihood_poisson,
                            glm_likelihood_gaussian)
from .utils import one_hot_decode, decision_boundary_1D


def top_20p(y_true, y_score):
    '''
    Measures the percent of true positives that are captured in the top 20%
    of the population (sorted by predicted probabilities).

    Parameters
    ----------
    y_true : numpy array
        True values (observed response).
    y_score : numpy array
        Predicted probabilities.

    Returns
    -------
    top_20 : float
        Percent of true positives that are captured in the top 20% 
        of the population.

    '''
    if isinstance(y_true, pd.DataFrame) or isinstance(y_true, pd.Series):
        y_true = y_true.values
    cutoff = int(0.2*y_true.shape[0])
    prob_sort = np.argsort(y_score)[::-1]
    top_20 = y_true[prob_sort][:cutoff].sum()/y_true.sum()
    return top_20


# For use with custom scorers in sklearn, otherwise, use top_20p
def top_20p_score(estimator, X, y, yprob=None):
    '''
    Measures the percent of true positives that are captured in the top 20%
    of the population (sorted by predicted probabilities).
    
    For use with custom scorers in sklearn, otherwise, use top_20p.

    Parameters
    ----------
    estimator : sklearn model, or similar
        The model. Needs to have a predict_proba method.
    X : numpy array or pandas dataframe
        The design or feature matrix.
    y : numpy array or pandas series
        The target or response variable.
    yprob : None
        For compatibility with sklearn custom scorers. Default is None.

    Returns
    -------
    top_20 : float
        Percent of true positives that are captured in the top 20% 
        of the population.

    '''
    if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
        y_true = y.values
    else:
        y_true = y
    y_score = estimator.predict_proba(X)[:, 1]
    cutoff = int(0.2*y_true.shape[0])
    prob_sort = np.argsort(y_score)[::-1]
    top_20 = y_true[prob_sort][:cutoff].sum()/y_true.sum()
    return top_20


def auc_score(estimator, X, y):
    '''
    Area under the ROC curve. Wrapper for compatibility with sklearn
    custom scorers.

    Parameters
    ----------
    estimator : sklearn model, or similar
        The model. Needs to have a predict_proba method.
    X : numpy array or pandas dataframe
        The design or feature matrix.
    y : numpy array or pandas series
        The target or response variable.

    Returns
    -------
    auc : float
        The area under the roc curve.

    '''
    if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
        y_true = y.values
    else:
        y_true = y
    y_score = estimator.predict_proba(X)[:, 1]
    auc = roc_auc_score(y_true, y_score, average='weighted')
    return auc


def cox_snell_r2(ll_est, ll_null, n_obs):
    '''
    Cox-Snell Pseudo R-squared.

    Parameters
    ----------
    ll_est : numpy array
        Estimated Log likelihood.
    ll_null : numpy array
        Null-model log-likelihood.
    n_obs : int
        Number of observations.

    Returns
    -------
    cs_r2 : float
        Cox-Snell Pseudo R-squared.

    '''
    ratio = 2/n_obs
    cs_r2 = 1 - np.exp(ratio*(ll_null - ll_est))
    return cs_r2


def nagelkerke_r2(ll_null, n_obs, cs_r2=None, ll_est=None):
    '''
    Nagelkerke Pseudo R-squared. Provides a correction to Cox-Snell Pseudo
    R-squared to bound between 0, 1.

    Parameters
    ----------
    ll_null : numpy array
        Null-model log-likelihood.
    n_obs : int
        Number of observations.
    cs_r2 : float
        Cox-Snell Pseudo R-squared. Default is None.
    ll_est : numpy array
        Only needed if cs_r2 not provided. Estimated Log likelihood.
        Default is None.

    Returns
    -------
    cs_r2 : float
        Nagelkerke Pseudo R-squared.

    '''
    ratio = 2/n_obs
    r2_max = 1 - np.exp(ratio*ll_null)
    if cs_r2 is None:
        cs_r2 = 1 - np.exp(ratio*(ll_null - ll_est))
    # Correction to CS R2 to bound between 0, 1
    n_r2 = cs_r2/r2_max
    return n_r2


def tjur_r2(y_true, y_score):
    '''
    Tjur Pseudo R-squared

    Parameters
    ----------
    y_true : numpy array
        True values (observed response).
    y_score : numpy array
        Predicted probabilities.

    Returns
    -------
    t_r2 : float
        Tjur Pseudo R-squared.

    '''
    y_mu1 = y_score[y_true == 1].mean()
    y_mu0 = y_score[y_true == 0].mean()
    t_r2 = np.abs(y_mu1 - y_mu0)
    return t_r2


def mcfadden_r2(ll_est, ll_null):
    '''
    McFadden's Pseudo R-squared when the saturated model is not available.

    Parameters
    ----------
    ll_est : numpy array
        Estimated Log likelihood.
    ll_null : numpy array
        Null-model log-likelihood.

    Returns
    -------
    m_r2 : float
        McFadden Pseudo R-squared.

    '''
    m_r2 = 1 - (ll_est/ll_null)
    return m_r2


def number_of_nonzero_coef(X, model):
    '''
    Calculates the number of non-zero coefficients in a model. Useful for
    evaluating models with an L1 penalty.

    Parameters
    ----------
    X : numpy array or pandas dataframe
        The design or feature matrix.
    model : sklearn object, or similar
        The model, which needs to have a coef_ attribute.

    Returns
    -------
    num_coef : int
        Number of non-zero coefficients.

    '''
    num_coef = None
    try:
        num_coef = sum((True if c != 0 else False for c in model.coef_.reshape(-1,)))
    except:
        num_coef = X.shape[1]
    return num_coef


def conf_mat_metrics(y_true, y_pred, conf_metric='all'):
    '''
    Confusion Matrix metrics for 0/1 class classification. Assumes 1 is the
    positive class (i.e. 1 == true positive, 0 == true negative).

    Parameters
    ----------
    y_true : numpy array
        True values (observed response).
    y_pred : numpy array
        Predicted 0/1 values.
    conf_metric : str, optional
        The metrics to return, can be `Sens/Recall`, `Specificity`, `ppv`,
        `npv`, `conf_mat`, or `all`. The default is 'all'.

    Returns
    -------
    dict, or float
        Returns dictionary for `all`, otherwise, float for the metric of
        interest.

    '''
    confMat = confusion_matrix(y_true, y_pred)
    metrics = {'Sens/Recall': 0,
               'Specificity': 0,
               'ppv': 0,
               'npv': 0,
               'Conf_Mat': 0}
    tn = confMat[0, 0]
    tp = confMat[1, 1]
    fn = confMat[0, 1]
    fp = confMat[1, 0]
    tnfp = tn + fp
    tpfn = tp + fn
    tpfp = tp + fp
    tnfn = tn + fn
    if tnfp != 0:
        metrics['Specificity'] = tn/tnfp
    if tpfn != 0:
        metrics['Sens/Recall'] = tp/tpfn
    if tpfp != 0:
        metrics['ppv'] = tp/tpfp
    if tnfn != 0:
        metrics['npv'] = tn/tnfn
    if tpfp*tpfn*tnfp*tnfn != 0:
        metrics['matthews_coef'] = ((tp*tn) - (fp*fn))/np.sqrt(tpfp*tpfn*tnfp*tnfn)
    metrics['Conf_Mat'] = confMat
    
    if conf_metric == 'all':
        return metrics
    else:
        return metrics[conf_metric]     

    
def bias(y_true, y_pred):
    '''
    Calculate model bias (systemic errors in estimation) in Linear Regression.
    
    Defined as The average distance (and direction) the predictions are from
    the true values (mean(y_pred - y_true)). If bias < 0, predictions are too
    low, and if bias > 0, then predictions are too high. Bias is measured in 
    the same units as the response.

    Parameters
    ----------
    y_true : numpy array
        True values (observed response).
    y_pred : numpy array
        Predicted values.

    Returns
    -------
    bias : float
        Model bias.

    '''
    ydiff = y_pred - y_true
    bias = np.mean(ydiff)
    return bias


def rpmse(y_true, y_pred, root=True):
    '''
    Calculate model PMSE/RPMSE (root predictive mean squared error) in
    Linear Regression.
    
    PMSE or MSE is defined as the average squared distance the predictions
    are from the true values (mean((y_pred - y_true)^2)), and measure how far 
    off the predictions are on average (i.e. how variable the predictions are).
    The root MSE is convenient because it is measured in the same units as the
    response.

    Parameters
    ----------
    y_true : numpy array
        True values (observed response).
    y_pred : numpy array
        Predicted values.

    Returns
    -------
    bias : float
        Model bias.

    '''
    ydiff = y_pred - y_true
    if root:
        return np.sqrt(np.mean(ydiff**2))
    else:
        return np.mean(ydiff**2)


def r2(y_true, y_pred):
    '''
    R-squared for a linear regression model. The percent of the variance in 
    the response that can be explained by the predictors.
    
    SSE : Sum of Squares Errors (Residuals)
    SSR : Sum of Squares Regression
    SST : Sum of Squares Total
    
    Parameters
    ----------
    y_true : numpy array
        True values (observed response).
    y_pred : numpy array
        Predicted values.

    Returns
    -------
    rsquared : float
        R-squared.

    '''
    y_bar = np.mean(y_true)
    SST = np.sum((y_true - y_bar)**2)
    SSE = np.sum((y_true - y_pred)**2)
    #SSR = np.sum((y_pred - y_bar)**2)
    rsquared = 1 - SSE/SST
    # Note: the following deifinitions are only equal to the above when the
    # model is linear, and will only be equivalent for the training data (i.e.
    # the test data is not guarenteed to have equivalent definitions of R2)
    # The equivalence of these definitions will hold for some non-linear
    # models, but not for others. Therefore,
    # Kvalseth TO. Cautionary Note About R2. Am. Statistic. 1985;39:279–285.
    # recommends using 1 - SSE/SSR for the general cases because by it's 
    # definition, it is the most robust to a general suite of problems
    
    # alternative R2: np.corrcoef((ypred, ytrue))**2
    # rsquared = SSR/SST
    return rsquared


def lr_se_fromModel(estimator, X, y):
    '''
    Linear Regression Model Standard deviation (error) estimate, or,
    the estimate of the true standard deviation of the underlying distribution
    for y_true. This is the maximum likelihood estimate for sigma, where
    y_true ~ N(y_pred, sigma^2).

    Parameters
    ----------
    y_true : numpy array
        True values (observed response).
    y_pred : numpy array
        Predicted values.
    num_params : int
        Number of Model Parameters

    Returns
    -------
    sigma : float
        Model error estimate.

    '''
    p = X.shape[1]
    y_pred = estimator.predict(X)
    sigma = np.sqrt(np.sum((y - y_pred)**2)/(y.shape[0] - p - 1))
    return sigma


def lr_se_fromPrediction(y_true, y_pred, num_params=2):
    '''
    Linear Regression Model Standard deviation (error) estimate, or,
    the estimate of the true standard deviation of the underlying distribution
    for y_true. This is the maximum likelihood estimate for sigma, where
    y_true ~ N(y_pred, sigma^2).

    Parameters
    ----------
    estimator : sklearn model, or similar
        The model. Needs to have a predict_proba method.
    X : numpy array or pandas dataframe
        The design or feature matrix.
    y : numpy array or pandas series
        The target or response variable.

    Returns
    -------
    sigma : float
        Model error estimate.

    '''
    sigma = np.sqrt(np.sum((y_true - y_pred)**2)/(y_true.shape[0] - num_params - 1))
    return sigma


def adj_r2(y_true, y_pred, X, rsquared=None):
    '''
    Adjusted R-squared for a linear regression model. The percent of the
    variance in the response that can be explained by the predictors. The 
    adjustment is for the ratio or predictors to observations.

    Parameters
    ----------
    y_true : numpy array
        True values (observed response).
    y_pred : numpy array
        Predicted values.
    X : numpy array
        The design or feature matrix.
    rsquared : float
        The Unadjusted R-squared for the model, if available. Otherwise,
        it's calculated from `r2`

    Returns
    -------
    rsquared : float
        R-squared.

    '''
    if rsquared is None:
        rsquared = r2(y_true, y_pred)
    adj_r2 = 1 - (1 - rsquared)*((X.shape[0] - 1)/(X.shape[0] - X.shape[1] - 1))
    return adj_r2
    

def aicCalc(loglike, num_model_params, sample_size, c=2, metric="aicc"):
    '''
    Generic Akaike information criterion (AIC) calculation for models.
    Includes metrics for AIC: `aic`, Corrected AIC: `aicc`, Bayesian
    Information Criterion (BIC): `bic`, or extended BIC: `ebic`.

    Parameters
    ----------
    loglike : float
        The loglikelihood of the model.
    num_model_params : int
        Number of parameters in the model.
    sample_size : int
        Number of observations.
    c : int or float, optional
        correction parameter for ebic. The default is 2.
    metric : str, optional
        The metric to evaluate (see main description). The default is "aicc".

    Raises
    ------
    ValueError
        Raised if no valid metric is selected.

    Returns
    -------
    ic : float
        The IC value.

    '''
    ic = None
    if metric == "aic" or metric == "aicc":
        ic = (2*num_model_params) - 2*(loglike)
        if metric == "aicc":
            ic = (ic + ((2*(num_model_params**2) +
                    2*num_model_params) /
                   (sample_size - num_model_params - 1))
                if (sample_size - num_model_params - 1) != 0 else ic)
    elif metric == "bic":
        ic = (np.log(sample_size)*num_model_params) - 2*loglike
    elif metric == "ebic":
        ic = (c*np.log(sample_size)*num_model_params) - 2*loglike
    if ic is not None:
        return ic
    else:
        raise ValueError("Not a valid method. No criteria was calculated")
        

def glm_regularized_AIC(X, Y, reg_mod, unreg_mod,
             tol=1e-6, method="kawano", family="binomial"):
    '''
    Calculate AIC for a Generalized Linear Model with regularization.
    
    See 'AIC for the Lasso in GLMs', Y. Ninomiya and S. Kawano (2016)

    Parameters
    ----------
    X : numpy array or pandas dataframe
        Feature or design matrix.
    Y : numpy array or pandas series
        Target or response variable.
    reg_mod : sklearn, or similar
        The regularized model.
    unreg_mod : sklearn, or similar
        The unregularized model.
    tol : float, optional
        Tolerance cutoff for counting non-zero coefficients. The default is
        1e-6.
    method : str, optional
        The method for calculating the AIC. Either `kawano` or `Hastie`.
        The default is "kawano".
    family : str, optional
        The type of generalised linear model. The default is "binomial".

    Raises
    ------
    ValueError
        Raised if an invalid family is picked.

    Returns
    -------
    aic : float
        The calculated AIC.

    '''
    # requires predict_proba method for logreg, and predict method for others, for poisson, predict output should be lambda, i.e. it should already be exponentiated
    if isinstance(X, pd.DataFrame):
        X = X.values
    if isinstance(Y, pd.DataFrame or pd.Series):
        Y = Y.values
    aic = None
    if family == "binomial":
        nllf = glm_likelihood_bernoulli
        unreg_prob = unreg_mod.predict_proba(X)[:, 1]
        reg_prob = reg_mod.predict_proba(X)[:, 1]
        y_mat_unreg = np.diag(unreg_prob * (1 - unreg_prob))
        y_mat_reg = np.diag(reg_prob * (1 - reg_prob))
    elif family == "poisson":
        nllf = glm_likelihood_poisson
        unreg_pred = unreg_mod.predict(X)
        reg_pred = reg_mod.predict(X)
        y_mat_unreg = np.diag(unreg_prob)
        y_mat_reg = np.diag(reg_prob)
    elif family == "gaussian":
        nllf = glm_likelihood_gaussian
        unreg_pred = unreg_mod.predict(X)
        reg_pred = reg_mod.predict(X)
        y_mat_unreg = np.diag(unreg_pred)
        y_mat_reg = np.diag(reg_pred)
    else:
        raise ValueError("Not a valid family")
    
    reg_mod_coef = np.concatenate((reg_mod.intercept_, np.squeeze(reg_mod.coef_)))
    nonzero_idx = np.where([True if np.abs(coef) > tol else False for coef in reg_mod_coef])[0]
    if method == "kawano":
        X_nz = add_constant(X[:, nonzero_idx], prepend=True)
        
        j22 = np.linalg.multi_dot([X_nz.T, y_mat_reg, X_nz])
        j22_2 = np.linalg.multi_dot([X_nz.T, y_mat_unreg, X_nz])
        
        negloglike = nllf(reg_mod_coef, X, Y, lamb=0, l_norm=0)
        aic = 2*negloglike + np.sum(np.diag(np.linalg.inv(j22).dot(j22_2)))
    else:
        # Tibshirani, Hastie, Zou 2007, On the degrees of freedom on the lasso
        negloglike = nllf(reg_mod_coef, X, Y, lamb=0, l_norm=0)
        # Not 100% sure on this calculation. should it be 2*len(nonzero_idx),
        #the count of nonzero columns, or 2*mean(non_zero_idx)?
        aic = 2*negloglike + 2*len(nonzero_idx)
    return aic


def js_div(px, py):
    '''
    Jensen-Shannon Divergence, which is a smoothed version of KL divergence.
    
    px: Probability of x (float or array of floats)
    py: Probability of y (float or array of floats)
    '''
    midpoint = (px + py)*0.5
    js = rel_entr(px, midpoint)*0.5 + rel_entr(py, midpoint)*0.5
    return np.sum(js)


def kl_div(px, py):
    '''
    KL divergence.
    
    Note: scipy has a KL divergence function of the same name, but it adds
    extra terms. 
    
    px: Probability of x (float or array of floats)
    py: Probability of y (float or array of floats)
    '''
    kl = np.sum(px*np.log(px/py))
    return kl


def HellingerDistanceMVN(mu1, mu2, cov1, cov2, squared=False):
    '''
    Quantifies the similarity between two multivariate normal distributions.

    Parameters
    ----------
    mu1 : numpy array
        Mean of the first distribution.
    mu2 : numpy array
        Mean of the second distribution.
    cov1 : numpy array
        Covariance of the first distribution.
    cov2 : numpy array
        Covariance of the second distribution.
    squared : bool, optional
        Return the squared distance. The default is False.

    Returns
    -------
    float
        Hellinger Distance.

    '''
    mu_diff = mu1 - mu2
    det_cov1 = np.linalg.det(cov1)
    det_cov2 = np.linalg.det(cov2)
    cov_sum = (cov1 + cov2)/2
    det_cov_sum = np.linalg.det(cov_sum)
    inv_cov_sum = np.linalg.inv(cov_sum)
    
    dets = (((det_cov1**0.25)*(det_cov2**0.25))/det_cov_sum**0.5)
    expon = -0.125*mu_diff.dot(inv_cov_sum).dot(mu_diff)

    H2 = 1 - dets*np.exp(expon)
    if squared:
        return H2
    else:
        return np.sqrt(H2)

def HellingerDistanceUN(mu1, mu2, sd1, sd2, squared=False):
    '''
    Quantifies the similarity between two Univariate normal distributions.

    Parameters
    ----------
    mu1 : float
        Mean of the first distribution.
    mu2 : float
        Mean of the second distribution.
    sd1 : float
        Standard Deviation of the first distribution.
    sd2 : float
        Standard Deviation of the second distribution.
    squared : bool, optional
        Return the squared distance. The default is False.

    Returns
    -------
    float
        Hellinger Distance.

    '''
    mu_diff2 = (mu1 - mu2)**2
    sd_sum = (sd1**2 + sd2**2)
    H2 = 1 - np.sqrt((2*sd1*sd2)/sd_sum)*np.exp(-0.25*(mu_diff2/sd_sum))
    
    if squared:
        return H2
    else:
        return np.sqrt(H2)
    

def HotellingsTwoSampleMVTtest(mu1, mu2, cov1, cov2, n1, n2, pval=True):
    '''
    Used for multivariate hypothesis testing. Can test the similarity of two
    Multivariate samples assumed to follow a normal distribution.

    Parameters
    ----------
    mu1 : numpy array
        Mean of the first distribution.
    mu2 : numpy array
        Mean of the second distribution.
    cov1 : numpy array
        Covariance of the first distribution.
    cov2 : numpy array
        Covariance of the second distribution.
    n1 : int
        Number of observations in the first sample.
    n2 : int
        Number of observations in the first sample.
    pval : bool, optional
        Calculate the p-value of the statistic. The default is True.

    Returns
    -------
    float
        Either p-value or t-statistic.

    '''
    mu_diff = mu1 - mu2
    pooled_cov = (((n1 - 1)*cov1) + ((n2 - 1)*cov2))/(n1 + n2 - 2)
    inv_pooled_cov = np.linalg.inv(pooled_cov)
    
    t2 = ((n1*n2)/(n1 + n2))*mu_diff.dot(inv_pooled_cov).dot(mu_diff)
    if pval:
        # the T2 dist is proportional to an F-dist. For simplicity...
        p = cov1.shape[0]
        m = (n1 + n2 - p - 1)
        
        f = (m/((n1 + n2 - 2)*p))*t2
        pv = f.sf(f, dfn=p, dfd=m) 
        return pv
    else:
        return t2
    

def accuracy(y_true, y_pred):
    if y_true.shape[1] != 1:
        return np.mean(one_hot_decode(y_true) == one_hot_decode(y_pred))
    else:
        return np.mean(y_true == decision_boundary_1D(y_pred))



import numpy as np

from ..metrics import rpmse, bias, accuracy
from ..tmath import norm, norm_der
from .activations import ( relu, elu,
                     sigmoid, sigmoid_der,
                     softmax, softmax_der, relu_der, cross_entropy,
                     cross_entropy_der, selu, selu_der, elu_der,
                     softmax_cross_entropy_der,
                     softmax_cross_entropy_der_fullmath,
                     sigmoid_cross_entropy_binary_der,
                     cross_entropy_binary, cross_entropy_binary_der,
                     linear, mse, mse_linear_der)
from .initializers import (he_uniform, he_normal, xavier_uniform,
                             xavier_normal, lecun_uniform, lecun_normal,
                             random
                             )
from .optimizers import gradient_descent, adam


class NeuralNetwork(object):
    def __init__(self,
                 model,
                 eval_size=10,
                 batch_size=64,
                 num_iterations=500,
                 optimizer='adam',
                 optimizer_args={'learning_rate': 0.001,
                                 'beta1': 0.9,
                                 'beta2': 0.999,
                                 'eps': 1e-8},
                 m_scale=1,
                 bn_tol=1e-9,
                 bn_momentum=0,
                 scorer='accuracy',
                 shuffle=False,
                 print_cost=True,
                 random_state=42):
        '''
        The constructor for the NeuralNetwork class.

        Parameters
        ----------
        model : dict
            A dictionary containing the model components, layers, and other
            specifications. The dictionary should have the following general
            structure:
                {
                 'hidden0': {'depth': 10,
                             'activation': 'relu',
                             'derivative': 'relu_der',
                             'activation_args': {},
                             'initializer': 'he_uniform',
                             'dropout_keep_prob': 1,
                             'lambda': {'Weight': 0,
                                       'activity': 0,
                                       'bias': 0
                                      },
                            'lp_norm': {'Weight': 2,
                                       'activity': 2,
                                       'bias': 2
                                      },
                            'use_batch_norm': False
                             },
                 'output': {'activation': 'softmax',
                            'activation_args': {},
                            'cost': 'cross_entropy',
                            'cost_args': {},
                            'derivative': 'softmax_cross_entropy_der',
                            'derivative_args': {},
                            'initializer': 'xavier_normal',
                            'evaluation_metric': 'accuracy',
                            'lambda': {'Weight': 0,
                                       'activity': 0,
                                       'bias': 0
                                      },
                            'lp_norm': {'Weight': 2,
                                       'activity': 2,
                                       'bias': 2
                                      },
                            'use_batch_norm': False
                            }
                 }
            Each layer should have the components defined above, however, not
            every component needs to be used (for example, setting
            dropout_keep_prob = 1 disables dropout). There can be as many
            hidden layers as desired (including none). Simply copy the
            'hidden1' sub-dictionary before the output layer to add a new
            hidden layer. However, the network must have an output layer
            defined. The key names for the layers can be anything, but the
            output layer must be positioned last.
            
            A description of each layer key is defined below:

                activation (str or function): The activation function to be
                                              used. If custom function, it
                                              will pass the affine
                                              transformation of the current 
                                              layer as the first input to the
                                              function.
                activation_args (dict) : An optional dictionary for passing
                                         additional arguments to the activation
                                         or derivative function. If there are
                                         none to pass, use an empty dictionary.
                                         For hidden layers, the derivative and
                                         activation arguments should be the
                                         same, so they share this dictionary.
                cost (str or function): The cost function to be
                                        used. If custom function, it
                                        will pass the true Y values and
                                        the predicted Y values as the first
                                        two inputs to the function.
                cost_args (dict) : An optional dictionary for passing
                                   additional arguments to the 
                                   cost function. If there are
                                   none to pass, use an empty dictionary.
                                   Only applies ot the output layer.
                depth (int): The number of hidden nodes in the layer
                derivative (str or function): The derivative of the combined
                                              cost and output layer activation
                                              function to be
                                              used. If custom function, it
                                              will pass the true Y values,
                                              the predicted Y values, and the
                                              non-activated output layer values
                                              as the first inputs to the 
                                              function.
                derivative_args (dict) : An optional dictionary for passing
                                         additional arguments to the derivative
                                         function. If there are none to pass,
                                         use an empty dictionary. This only
                                         applies to the output layer.
                dropout_keep_prob (float) : The proportion of nodes to keep at
                                            the respective layer. Between 0
                                            and 1. If dropping 10% of the
                                            nodes, the keep prob is 0.9
                evaluation_metric (str or function) : An additional evaluation
                                                      metric to be used in 
                                                      training. This is only
                                                      used for printing an
                                                      additional output along
                                                      with cost at each epoch
                                                      or specified iteration to
                                                      track the training
                                                      progress
                initializer (str or function) : The function to be used in 
                                                initializing the layer weights
                                                and biases. If custom, it must
                                                accept two arguments,  
                                                'incoming' and 'outgoing',
                                                which represent how many inputs
                                                are recieved from the previous
                                                layer, and how many outputs
                                                will be calculated at the 
                                                current layer.
                lambda (dict) : A dictionary containing the regularization 
                                penalties for each type of regularization.
                                The options are:
                                    Weight (float) : The kernel or weight
                                                     regularizer
                                                     (recommended for use)
                                    activity (float) : A regularizer placed on
                                                       the activation function
                                                       output (experimental in
                                                       this code, not
                                                       recommended for use)
                                    bias (float) : A regularizer for the bias
                                                   (not recommended for use for
                                                   theoretical reasons, but
                                                   should be correct to use)
                                
                                A value of zero for any of the lambdas will
                                that regularization type for that layers.
                lp_norm (dict) : A dictionary containing the regularization 
                                 norm funcitons for each type of
                                 regularization.
                                 
                                 The options are:
                                    Weight (int) : The lp-norm for the weight
                                                   or kernel regularizer
                                    activity (int) : The lp-norm for the
                                                     activity regularizer
                                    bias (int) : The lp-norm for the bias
                                                 regularizer
                use_batch_norm (bool) : If true, perform batch normalization
                                        on the current layer. For this
                                        implementation, the batch norm layer
                                        is placed before the activation
                                        function and before dropout (if used
                                        together)
                
        eval_size : int, optional
            The number of model evaluations to perform before printing
            an output. It is recommended that this number be
            `int(n/batch_size) + sum([n % batch_size != 0])` where n is the
            number of observations
        batch_size : int, optional
            The number of observations used to update the model at each step.
            The default is 64.
        num_iterations : int, optional
            The total number of full passes through the data to perform
            (i.e. the number of epochs). The default is 500.
        optimizer : str, optional
            The type of optimizer to use for gradient descent.
            The default is 'adam'.
        optimizer_args : dict, optional
            Optional arguments to send to the optimizer (learning rate, etc.). 
            The default is {'learning_rate': 0.001,
                            'beta1': 0.9,
                            'beta2': 0.999,
                            'eps': 1e-8}.
        m_scale : float, optional
            An optional scaling paramter to scale up or down the cost and
            gradient values. The default is 1.
        bn_tol : float, optional
            The tolerance used in the batch norm equations.
            The default is 1e-9.
        bn_momentum : float, optional
            The momentum used in the exponential moving average for the mean
            and variance of the batch norm process. The default is 0.
        scorer : str or function, optional
            The function used in the score method. If custom, it must accept
            the true Y values and the predicted y values as the first two
            arguments of the function.
            The default is 'accuracy'.
        shuffle : bool, optional
            Shuffle the training set before training. The default is False.
        print_cost : bool, optional
            Print the cost (and possibly another metric) at each eval_step.
            The default is True.
        random_state : int, optional
            The random state of the process (for reproducibility).
            The default is 42.

        Returns
        -------
        None.

        '''
        self.model = model
        self.num_layers = len(model.keys())
        self.eval_size = eval_size
        self.batch_size = batch_size
        self.num_iterations = num_iterations
        self.optimizer_args = optimizer_args
        self.m_scale = m_scale
        self.bn_tol = bn_tol
        self.bn_momentum = bn_momentum
        self.print_cost = print_cost
        self.random_state = random_state
        self.shuffle = shuffle
        self.scorer = scorer
        self.activations = {'linear': linear,
                            'mse': mse,
                            'mse_linear_der': mse_linear_der,
                            'relu': relu,
                            'relu_der': relu_der,
                            'elu': elu,
                            'elu_der': elu_der,
                            'selu': selu,
                            'selu_dir': selu_der,
                            'sigmoid': sigmoid,
                            'sigmoid_der': sigmoid_der,
                            'softmax': softmax,
                            'softmax_der': softmax_der,
                            'cross_entropy': cross_entropy,
                            'cross_entropy_der': cross_entropy_der,
                            'cross_entropy_binary': cross_entropy_binary,
                            'cross_entropy_binary_der': cross_entropy_binary_der,
                            'softmax_cross_entropy_der': softmax_cross_entropy_der,
                            'softmax_cross_entropy_der_fullmath': softmax_cross_entropy_der_fullmath,
                            'sigmoid_cross_entropy_binary_der': sigmoid_cross_entropy_binary_der
                            }
        self.initializers = {'he_uniform': he_uniform,
                             'he_normal': he_normal,
                             'xavier_uniform': xavier_uniform,
                             'xavier_normal': xavier_normal,
                             'lecun_uniform': lecun_uniform,
                             'lecun_normal': lecun_normal,
                             'random': random
                             }
        self.optimizers = {'adam': adam,
                           'gradient_descent': gradient_descent}
        self.scorer = {'accuracy': accuracy,
                       'rpmse': rpmse,
                       'bias': bias
                       }
        self.optimizer = 'adam'
        self.update_parameters = self.optimizers[optimizer]
    

    def initialize_wb(self, X):
        '''
        Initialize the network.

        Parameters
        ----------
        X : numpy array
            The input data.

        Returns
        -------
        wb_list : dict
            A dictionary of the weights/biases for each layer.

        '''
        wb_list = {}
        xrows = X.shape[0]
        X = X.reshape(xrows, -1)
        prev_col = X.shape[1]
    
        for k, key in enumerate(self.model.keys()):
            cur_row = prev_col
            cur_col = self.model[key]['depth']
            
            init_func = self.initializers[self.model[key]['initializer']]
            
            # Weights cannot be initialized to zero, or the model can't train
            W = init_func(cur_row, cur_col)
            
            # Zero is a common bias initialization. Some argue that a small
            # positive value like 0.01 should be used instead. Others argue
            # that makes it worse. LSTMs typically initialize bias at 1
            b = np.zeros(shape=(1, cur_col))
            wb_list["Weight" + str(k)] = W
            wb_list["bias" + str(k)] = b
            if self.model[key]['use_batch_norm']:
                wb_list['gamma' + str(k)] = np.ones(shape=(1, cur_col))
                wb_list['beta' + str(k)] = np.zeros(shape=(1, cur_col))
            prev_col = cur_col
    
        return wb_list

    def forward_prop(self, X, wb, batch_norm, train=True, sample=False):
        '''
        The forward propagation step

        Parameters
        ----------
        X : numpy array
            The input data.
        wb : dict
            A dictionary of the weights/biases for each layer.
        batch_norm : dict
            A dictionary containing the initial (or previous) batch norm
            results.
        train : bool, optional
            Whether the forward propagation method is being used to train or
            calculate the network in it's current state.
            The default is True.
        sample : bool, optional
            Whether or not the forward propagation is being used to generate
            random smaples of the output from the distribution created using
            dropout. The default is False.

        Returns
        -------
        hz : numpy array
            The final predicted output.
        zs : dict
            A dictionary of the linearly (affine transform) activated values
            for each layer.
        batch_norm : dict
            A dictionary containing the initial (or previous) batch norm
            results.
        hzs : dict
            A dictionary of the fully activated values for each layer
        dropout : dict
            A dictionary of the dropout status for each layer.
        regularizers : dict
            A dictionary of the regularization status for each layer.

        '''
        hz = X
        zs = {}
        hzs = {}
        dropout = {}
        regularizers = {}
        
        for i, key in enumerate(self.model.keys()):
            indx = str(i)
            z = hz.dot(wb["Weight" + indx]) + wb["bias" + indx]
            
            # It is a highly debated topic whether or not batch norm or dropout
            # first, and on top of that, whether batch norm should occur before
            # or after activation. It likely depends on specific situations on
            # types of networks. Other sources say they shouldn't be used
            # together anyway. Since this is a demonstration, I chose to put
            # batch norm first, before acivation. Note: moving them around 
            # will affect the derivative, so if doing by hand and not using
            # autograd etc., watch out for this.
            
            if self.model[key]['use_batch_norm']:
                if train:
                    batch_norm['mu' + indx] = np.mean(z, axis=0)
                    batch_norm['var' + indx] = np.var(z, axis=0)
                    batch_norm['z_mu' + indx] = z - batch_norm['mu' + indx]
                    batch_norm['std' + indx] = np.sqrt(batch_norm['var' + indx] + self.bn_tol)
                    batch_norm['xhat' + indx] = batch_norm['z_mu' + indx]/batch_norm['std' + indx] 
                    batch_norm['norm_z' + indx] = wb['gamma' + indx]*batch_norm['xhat' + indx] + wb['beta' + indx]
                    
                    # Exponential running mean for mu/var, if desired. Use
                    # momentum = 0 for regular batch norm process
                    batch_norm['mu_r' + indx] = self.bn_momentum*batch_norm['mu_r' + indx] + (1 - self.bn_momentum)*batch_norm['mu' + indx]
                    batch_norm['var_r' + indx] = self.bn_momentum*batch_norm['var_r' + indx] + (1 - self.bn_momentum)*batch_norm['var' + indx]
                else:
                    batch_norm['xhat' + indx] = (z - batch_norm['mu_r' + indx])/np.sqrt(batch_norm['var_r' + indx] + self.bn_tol)
                    batch_norm['norm_z' + indx] = wb['gamma' + indx]*batch_norm['xhat' + indx] + wb['beta' + indx]
            else:
                batch_norm['norm_z' + indx] = z
            
            actf = self.activations[self.model[key]['activation']]
            hz = actf(batch_norm['norm_z' + indx], **self.model[key]['activation_args'])
            
            if i != self.num_layers - 1:
                if train or sample:
                    dropout_keep_prob = self.model[key]['dropout_keep_prob']
                else:
                    dropout_keep_prob = 1
                drop_mask = np.random.uniform(size=hz.shape)
                dropout["Weight" + indx] = drop_mask <= dropout_keep_prob
                hz = (hz*dropout["Weight" + indx])/dropout_keep_prob
                
            lamda_w = self.model[key]['lambda']['Weight']
            p_w = self.model[key]['lp_norm']['Weight']
            lamda_a = self.model[key]['lambda']['activity']
            p_a = self.model[key]['lp_norm']['activity']
            lamda_b = self.model[key]['lambda']['bias']
            p_b = self.model[key]['lp_norm']['bias']
            
            # kernel/weight regularizer
            # Note: information here is only recorded, it does not affect
            # the forward propagation calculations at all
            if train:
                regularizers["Weight" + indx] = (1/p_w)*lamda_w*norm(wb["Weight" + indx], p_w)
                regularizers["activity" + indx] = (1/p_a)*lamda_a*norm(hz, p_a)
                regularizers["bias" + indx] = (1/p_b)*lamda_b*norm(wb["bias" + indx], p_b)
                
            zs['z' + str(i)] = z
            hzs['hz' + str(i)] = hz
        return hz, zs, batch_norm, hzs, dropout, regularizers
        
    def back_prop(self, X, Y, wb, zs, batch_norm, hzs, dropout):
        '''
        The backward propagation step.

        Parameters
        ----------
        X : numpy array
            The input data.
        Y : numpy array
            The true Y values.
        wb : dict
            A dictionary of the weights/biases for each layer.
        zs : dict
            A dictionary of the linearly (affine transform) activated values
            for each layer.
        batch_norm : dict
            A dictionary containing the initial (or previous) batch norm
            results.
        hzs : dict
            A dictionary of the fully activated values for each layer
        dropout : dict
            A dictionary of the dropout status for each layer.

        Returns
        -------
        dwdb : dict
            A dictionary of the gradients with respect to the weights and
            biases.

        '''
        dwdb = {}
        batch_m = X.shape[0]
        keys = list(self.model.keys())
        for i in range(self.num_layers - 1, -1, -1):
            lamda_w = self.model[keys[i]]['lambda']['Weight']
            p_w = self.model[keys[i]]['lp_norm']['Weight']
            lamda_a = self.model[keys[i]]['lambda']['activity']
            p_a = self.model[keys[i]]['lp_norm']['activity']
            lamda_b = self.model[keys[i]]['lambda']['bias']
            p_b = self.model[keys[i]]['lp_norm']['bias']
            
            if i == self.num_layers - 1:
                dcostoutf = self.activations[self.model[keys[i]]['derivative']]
                dZ = dcostoutf(Y, hzs["hz" + str(i)], zs["z" + str(i)],
                               **self.model[keys[i]]['derivative_args'])/(batch_m*self.m_scale)
                dZ += lamda_a*norm_der(hzs["hz" + str(i)], p_a)/(batch_m*self.m_scale)
                
                # Batchnorm step, if applicable
                if self.model[keys[i]]['use_batch_norm']:
                    dxhat = dZ * wb["gamma" + str(i)]
                    dvar = -0.5*np.sum(dxhat*batch_norm["z_mu"+str(i)], axis=0)*(1/batch_norm["std"+str(i)]**3)
                    dxdstd = dxhat/batch_norm["std"+str(i)]
                    dmu = -np.sum(dxdstd, axis=0) 
                    dmu -= 2*dvar*np.mean(batch_norm["z_mu"+str(i)], axis=0)
                    
                    dgamma = np.sum(dZ*batch_norm["xhat"+str(i)], axis=0)
                    dbeta = np.sum(dZ, axis=0)
                    
                    dwdb["gamma" + str(i)] = dgamma
                    dwdb["beta" + str(i)] = dbeta
                    
                    dZ = dxdstd + 2*dvar*batch_norm["z_mu"+str(i)]/zs["z"+str(i)].shape[0]
                    dZ += dmu/zs["z"+str(i)].shape[0]

            else:
                dZ = dZ.dot(wb["Weight" + str(i + 1)].T)
                
                dropout_keep_prob = self.model[keys[i]]['dropout_keep_prob']
                dZ = dZ * dropout["Weight" + str(i)]/dropout_keep_prob
                
                dactf = self.activations[self.model[keys[i]]['derivative']]

                dZ = dZ * dactf(zs["z" + str(i)],
                                **self.model[keys[i]]['activation_args'])
                dZ = dZ + lamda_a*norm_der(hzs["hz" + str(i)], p_a)/(batch_m*self.m_scale)
                
                # Batchnorm step, if applicable
                if self.model[keys[i]]['use_batch_norm']:
                    dxhat = dZ * wb["gamma" + str(i)]
                    dvar = -0.5*np.sum(dxhat*batch_norm["z_mu"+str(i)], axis=0)*(1/batch_norm["std"+str(i)]**3)
                    dxdstd = dxhat/batch_norm["std"+str(i)]
                    dmu = -np.sum(dxdstd, axis=0) 
                    dmu -= 2*dvar*np.mean(batch_norm["z_mu"+str(i)], axis=0)
                    
                    dgamma = np.sum(dZ*batch_norm["xhat"+str(i)], axis=0)
                    dbeta = np.sum(dZ, axis=0)
                    
                    dwdb["gamma" + str(i)] = dgamma
                    dwdb["beta" + str(i)] = dbeta
                    
                    dZ = dxdstd + 2*dvar*batch_norm["z_mu"+str(i)]/zs["z"+str(i)].shape[0]
                    dZ += dmu/zs["z"+str(i)].shape[0]
                
            if i == 0:
                A = X
            else:
                A = hzs["hz" + str(i - 1)]
                
            # m doesn't come from the derivative. We are dividing by m because we
            # technically calculated the gradient for each observation. Thus, we
            # average them to get the overall impact. We could also just sum and 
            # not use the mean, but having a lower value is helpful to not run
            # into errors with large numbers
            
            dW = (batch_m*self.m_scale)*(np.dot(A.T, dZ) + lamda_w*norm_der(wb["Weight" + str(i)], p_w)/p_w)
            dB = (batch_m*self.m_scale)*(np.sum(dZ, axis=0, keepdims=True) + lamda_b*norm_der(wb["bias" + str(i)], p_b)/p_b)
            dwdb["Weight" + str(i)] = dW
            dwdb["bias" + str(i)] = dB
        return dwdb


    def fit(self, X, Y):
        '''
        Fits the Neural Network.

        Parameters
        ----------
        X : numpy array
            The input data.
        Y : numpy array
            The true Y values.

        Returns
        -------
        self
            The fitted model.

        '''
        # This is only here for debugging or reproducibility
        if self.random_state is not None:
            np.random.seed(self.random_state)

        # Store total training observations
        self.m = X.shape[0]
        
        # Reshape Y so that it is a matrix, even in 1D
        Y = Y.reshape(self.m, -1)
        self.model['output']['depth'] = Y.shape[1]
        
        # Check for an additional metric to evaluate other than cost
        output_layer_components = list(self.model['output'].keys())
        has_metric = False
        if 'evaluation_metric' in  output_layer_components:
            if isinstance(self.model['output']['evaluation_metric'], str):
                met_func = self.scorer[self.model['output']['evaluation_metric']]
            else:
                met_func = self.model['output']['evaluation_metric']
            has_metric = True
        
        # Initialize the graph
        wb = self.initialize_wb(X)
        
        # Total batches needed for minibatch
        if self.batch_size > self.m:
            total_batches = self.m
        else:
            total_batches = int(self.m/self.batch_size) + sum([self.m % self.batch_size != 0])
        
        # Initialize result arrays and weights
        costs = np.zeros(self.num_iterations*total_batches)
        metric = np.zeros(self.num_iterations*total_batches)
        
        batch_norm = {'mu_r'+str(i): 0 for i in range(self.num_layers)}
        batch_norm.update({'var_r'+str(i): 1 for i in range(self.num_layers)})
        
        batches = [(b*self.batch_size, (b+1)*self.batch_size) for b in range(total_batches)]
        
        # TODO: Figure out if this (and below) is the proper way to do minibatch
        if self.shuffle:
            rand_indx = np.random.choice(range(X.shape[0]), X.shape[0])
            X = X[rand_indx, :]
            Y = Y[rand_indx, :] 

        # marker for counting total iterations in minibatch        
        count = 0
        for i in range(self.num_iterations):
            # TODO: figure out if each batch is meant to be random in minibatch
            #batch_order = np.random.choice(range(total_batches), total_batches)
            batch_order = range(total_batches)
            for j in batch_order:
                
                X_batch = X[batches[j][0]:batches[j][1], :]
                Y_batch = Y[batches[j][0]:batches[j][1], :]
                
                batch_m = X_batch.shape[0]
                
                # Forward propagation. 
                (hz, zs, batch_norm, hzs, dropout, regularizers) = self.forward_prop(X_batch,
                                                                                     wb,
                                                                                     batch_norm,
                                                                                     train=True,
                                                                                     sample=False)
                
                # Calculate cost without regularization
                costf = self.activations[self.model['output']['cost']]
                cost = costf(Y_batch, hz,
                             **self.model['output']['cost_args'])
                
                # Get regularization information for the cost function,
                # the activity, weight, and bias regularizers
                norms_w = np.sum([regularizers["Weight"+str(l)] for l in range(self.num_layers)])
                norms_a = np.sum([regularizers["activity"+str(l)] for l in range(self.num_layers)]) 
                norms_b = np.sum([regularizers["bias"+str(l)] for l in range(self.num_layers)]) 
                
                # Update Cost with regularization
                cost = (cost + norms_w + norms_a + norms_b)/(batch_m*self.m_scale)
                
                # Backpropagation.
                dwdb = self.back_prop(X_batch, Y_batch, wb, zs, 
                                      batch_norm, hzs, dropout)
                
                # Update parameters with optimizing function
                if self.optimizer == 'adam':
                    if i == 0:
                        self.optimizer_args['mt'] = {}
                        self.optimizer_args['vt'] = {}
                        for key in wb.keys():
                            self.optimizer_args['mt'][key] = np.zeros(wb[key].shape)
                            self.optimizer_args['vt'][key] = np.zeros(wb[key].shape)
                    self.optimizer_args['i'] = count
                
                # Update weights, biases, and other parameters for batch norm
                wb = self.update_parameters(wb, dwdb, self.optimizer_args)
                
                # Store cost and other metric, if applicable
                costs[count] = cost
                if has_metric:
                    metric[count] = met_func(Y_batch, hz)
    
                # Print the cost/metric every eval_size iterations
                if self.print_cost and (count + 1) % self.eval_size == 0:
                    if has_metric:
                        print("Evaluation %i: Cost: %f, Metric: %f" % (int((count + 1)/self.eval_size),
                                                                      costs[((count + 1)-self.eval_size):(count + 1)].mean(),
                                                                      metric[((count + 1)-self.eval_size):(count + 1)].mean()))
                    else:
                        print("Evaluation %i: Cost: %f" % (int((count + 1)/self.eval_size),
                                                          costs[((count + 1)-self.eval_size):(count + 1)].mean()))
                count += 1
                
        self.wb = wb
        self.dwdb = dwdb
        self.costs = costs
        self.batch_norm = batch_norm
        if has_metric:
            self.metric = metric
        return self
    
    def predict(self, X):
        '''
        Predict the Y values based on the input X.

        Parameters
        ----------
        X : numpy array
            The input data.

        Returns
        -------
        hz : numpy array
            The predicted Y values.

        '''
        hz, _, _, _, _, _ = self.forward_prop(X, self.wb, self.batch_norm,
                                           train=False, sample=False)
        return hz
    
    def draw_predictive_samples(self, X, n_samples=1000, n_outputs=1):
        '''
        Draws predictive samples for each observation from the distribution
        created by using dropout.
        
        The rows of the output represent the observations in X, and the columns
        of the output represent the number of samples drawn. If Y has multiple
        outputs or classes for a single observation, then the shape of the
        output will be (X.shape[0], Y.shape[1]*n_samples). For example, if
        there are four observations in X, two classes in Y, and you want three
        samples, then the output will be:
            
            [[Y11_c1, Y11_c2] [Y12_c1, Y12_c2], [Y13_c1, Y13_c2]
             [Y21_c1, Y21_c2] [Y22_c1, Y22_c2], [Y23_c1, Y23_c2]
             [Y31_c1, Y31_c2] [Y32_c1, Y32_c2], [Y33_c1, Y33_c2]
             [Y41_c1, Y41_c2] [Y42_c1, Y42_c2], [Y43_c1, Y43_c2]]

        Parameters
        ----------
        X : numpy array
            The input data.
        n_samples : int, optional
            The number of samples to draw. The default is 1000.
        n_outputs : int, optional
            The number of outputs or classes in Y. The default is 1.

        Returns
        -------
        draws : numpy array
            The sampled predicted values.

        '''
        draws = np.zeros(shape=(X.shape[0], n_samples*n_outputs))
        for i in range(n_samples):
            hz, _, _, _, _ = self.forward_prop(X, self.wb, self.batch_norm,
                                               train=True, sample=True)
            draws[:, (i*n_outputs):((i+1)*n_outputs)] = hz
        return draws
    
    def score(self, X, y):
        '''
        A method to score/test the model based on the inputs X, y. This can
        be a custom function or called with a string from something built-in.

        Parameters
        ----------
        X : numpy array
            The input data.
        y : numpy array
            The true Y values.

        Returns
        -------
        float or numpy array
            The score result for X, y.

        '''
        if isinstance(self.scorer, str):
            score_func = self.scorer_list[self.scorer]
        else:
            score_func = self.scorer
        y_score = self.predict(X, self.wb)
        return score_func(y, y_score)



import numpy as np
from scipy.stats import truncnorm


# He (uniform or normal) for relu variations
def he_uniform(incoming, outgoing):
    W = np.random.uniform(size=(incoming, outgoing),
                          low=-np.sqrt(6/(incoming)),
                          high=np.sqrt(6/(incoming)))
    return W


def he_normal(incoming, outgoing):
    W = truncnorm.rvs(size=(incoming, outgoing), a=-2, b=2,
                      scale=np.sqrt(2/(incoming)))
    return W


# Xavier (uniform or normal) for sigmoid/softmax/tanh
def xavier_uniform(incoming, outgoing):
    W = np.random.uniform(size=(incoming, outgoing),
                          low=-np.sqrt(6/(incoming + outgoing)),
                          high=np.sqrt(6/(incoming + outgoing)))
    return W


def xavier_normal(incoming, outgoing):
    W = truncnorm.rvs(size=(incoming, outgoing), a=-2, b=2,
                      scale=np.sqrt(2/(incoming + outgoing)))
    return W


def lecun_uniform(incoming, outgoing):
    W = np.random.uniform(size=(incoming, outgoing),
                          low=-np.sqrt(3/(incoming)),
                          high=np.sqrt(3/(incoming)))
    return W


# For selu
def lecun_normal(incoming, outgoing):
    W = truncnorm.rvs(size=(incoming, outgoing), a=-2, b=2,
                      scale=np.sqrt(1/(incoming)))
    return W
    

def random(incoming, outgoing):
    W = np.random.normal(size=(incoming, outgoing), scale=0.01)
    return W



import numdifftools as nd
import numpy as np
import warnings

from scipy.stats import norm


def signIntervals(fun, lower=-50, upper=50, step=0.01, fun_args=None):
    '''
    Returns the boundaries of all possible roots in a function. Works well
    with Brent root method.

    Parameters
    ----------
    fun : function
        The function to find the roots of.
    lower : float, optional
        The minimum value to investigate. The default is -50.
    upper : float, optional
        The maximum value to investigate. The default is 50.
    step : float, optional
        The interval between values in (lower, upper) to investigate.
        The default is 0.01.
    fun_args : dict
        Optional arguments to pass to fun.

    Returns
    -------
    bounds : list
        A list of tuples containing the upper and lower bounds of possible
        roots.

    '''
    if fun_args is None:
        fun_args = {}
    x = np.arange(lower, upper, step)
    steps = len(x)
    fx = np.zeros(steps)
    for i in range(steps):
        fx[i] = fun(x[i], **fun_args)

    curr = np.sign(fx[0])
    changeList = np.zeros(1)

    for j in range(steps):
        if np.sign(fx[j]) != curr:
            changeList = np.concatenate((changeList, np.array(j).reshape(1, )))
            curr = np.sign(fx[j])

    bounds = []
    for k in range(len(changeList[:-1])):
        bounds.append((x[int(changeList[k])], x[int(changeList[k + 1])]))

    return bounds


def Brent_1DRoot(fun, opt=None, print_=True, fun_args=None):
    '''
    Brent Method for Roots

    This function works well with signIntervals. For example, if the function
    "func" is x^2 - 5, the following code snippet will return both roots:
        bds = signIntervals(easy)
        for i in bds:
            Brent_1DRoot(easy, opt={'bounds': i})

    Parameters
    ----------
    fun : function
        The function to find the roots of.
    opt : dict, optional
        Optional arguments for the optimization. The default is None.
    print_ : bool, optional
        Print Results. The default is True.
    fun_args : dict
        Optional arguments to pass to fun.

    Raises
    ------
    ValueError
        Raised if there are invalid bounds, or no valid root.
    TypeError
        Raised if bounds is not a valid type.

    Returns
    -------
    final_res : dict
        A dictionary containing the results.

    '''
    if fun_args is None:
        fun_args = {}
    
    dom_tol = 1e-8
    maxiter = 10000
    bounds = [-10000.0, 10000.0]
    fun_tol = 1e-8
    if opt is not None:
        opt_keys = list(opt.keys())
        if 'dom_tol' in opt_keys:
            dom_tol = opt['dom_tol']
        if 'maxiter' in opt_keys:
            maxiter = opt['maxiter']
        if 'bounds' in opt_keys:
            bounds = opt['bounds']
        if 'fun_tol' in opt_keys:
            fun_tol = opt['fun_tol']

    try:
        if len(bounds) != 2:
            raise ValueError("Bounds option must be of length 2")
        if not isinstance(bounds, (tuple, list, np.ndarray)):
            raise TypeError("""must be array or list-like
                                (np.ndarray, tuple, or list""")
    except KeyError:
        bounds = [-10000.0, 10000.0]

    a = min(bounds)
    b = max(bounds)
    fa = fun(a, **fun_args)
    fb = fun(b, **fun_args)

    if fa*fb >= 0:
        raise ValueError("""Root is not bracketed, or root does not exist.
                          (try changing the interval)""")

    if abs(fa) < abs(fb):
        a, b = b, a
        fa, fb = fb, fa

    c = (float(a) + float(b))/2
    fc = float(fa)
    mflag = True
    cur_iter = 1
    d = float(c)

    while (abs(b - a) > dom_tol) and cur_iter < maxiter:
        if abs(fa - fc) > dom_tol and abs(fb - fc) > dom_tol:
            # inverse quadratic interpolation
            q1 = a*fb*fc / ((fa - fb)*(fa - fc))
            q2 = b*fa*fc / ((fb - fa)*(fb - fc))
            q3 = c*fa*fb / ((fc - fa)*(fc - fb))
            s = q1 + q2 + q3
        else:
            # secant method
            s = b - fb*((b - a)/(fb - fa))

        delta = abs(2*fun_tol*abs(b))
        if ((s < ((3*a + b)/4) and s > b) or
           (mflag and abs(s - b) >= abs(b - c)/2) or
           (not mflag and abs(s - b) >= abs(c - d)/2) or
           (mflag and abs(b - c) < abs(delta)) or
           (not mflag and abs(c - d) < abs(delta))):
            # bisection method
            s = (a + b)/2
            mflag = True
        else:
            mflag = False
        fs = fun(s, **fun_args)
        d = float(c)
        c = float(b)

        if fa*fs < 0:
            b = float(s)
            fb = float(fs)
        else:
            a = float(s)
            fa = float(fs)

        if abs(fa) < abs(fb):
            a, b = b, a
            fa, fb = fb, fa
        cur_iter += 1

    if cur_iter >= maxiter:
            warnings.warn('Warning: May not have found roots. Number of' +
                          'iterations exceeded maximum', Warning)

    final_res = {'root': a,
                 'Final Function Value': fa,
                 'Domain Tolerances': abs(b - a),
                 'Iterations': cur_iter,
                 'a:b:s': (a, b, s)}
    if print_:
        print("Inputs: \n", opt, "\n\nFinal Results\n", final_res)

    return final_res


def AdaptiveNMorBrent(fun, start=None, opt=None, print_=True, fun_args=None):
    '''
    Nelder-Mead Method and Brent method (if one-dimensional) for finding the
    minimum of a function.

    Parameters
    ----------
    fun : function
        The function to find the roots of.
    start : numpy array
        The starting values for the optimization. Must be equal in length to
        the number of parameters being optimized.
    opt : dict, optional
        Optional arguments for the optimization. The default is None.
    print_ : bool, optional
        Print Results. The default is True.
    fun_args : dict
        Optional arguments to pass to fun.

    Raises
    ------
    ValueError
        Raised if there are invalid bounds, or no valid root.
    TypeError
        Raised if bounds is not a valid type.

    Returns
    -------
    final_res : dict
        A dictionary containing the results.

    '''
    dom_tol = 1e-8
    maxiter = 10000
    bounds = [-10000.0, 10000.0]
    fun_tol = 1e-8
    if opt is not None:
        opt_keys = list(opt.keys())
        if 'dom_tol' in opt_keys:
            dom_tol = opt['dom_tol']
        if 'maxiter' in opt_keys:
            maxiter = opt['maxiter']
        if 'bounds' in opt_keys:
            bounds = opt['bounds']
        if 'fun_tol' in opt_keys:
            fun_tol = opt['fun_tol']
    

    if start is None or len(start) == 1:
        # Continue with Brent Method
        try:
            bounds = opt['bounds']
            if len(bounds) != 2:
                raise ValueError("Bounds option must be of length 2")
            if not isinstance(bounds, (tuple, list, np.ndarray)):
                raise TypeError("""must be array or list-like
                                (np.ndarray, tuple, or list)""")
        except KeyError:
            bounds = [-10000.0, 10000.0]
        # 1/(phi^2) where phi is the golden ratio
        golden = (1/((1 + 5**0.5) / 2))**2
        max_x = max(bounds)
        min_x = min(bounds)
        x = w = v = max_x
        fw = fv = fx = fun(x, **fun_args)
        delta = delta2 = 0
        cur_iter = 1

        while(cur_iter <= maxiter):
            mid = (min_x + max_x)/2
            f1 = (dom_tol * abs(x) + dom_tol)/4
            f2 = 2 * f1
            if abs(x - mid) <= (f2 - (max_x - min_x)/2):
                break

            if abs(delta2) > f1:
                r = (x - w) * (fx - fv)
                q = (x - v) * (fx - fw)
                p = (x - v) * q - (x - w) * r
                q = 2 * (q - r)
                if q > 0:
                    p = -p
                q = abs(q)
                td = delta2
                delta2 = delta
                if ((abs(p) >= abs(q * td / 2)) or
                   (p <= q * (min_x - x)) or
                   (p >= q * (max_x - x))):
                    delta2 = min_x - x if (x >= mid) else max_x - x
                    delta = golden * delta2
                else:
                    delta = p / q
                    u = x + delta
                    if (u - min_x) < f2 or (max_x - u) < f2:
                        delta = -abs(f1) if (mid - x) < 0 else abs(f1)
            else:
                delta2 = min_x - x if x >= mid else max_x - x
                delta = golden * delta2

            u = (x + delta if abs(delta) >= f1
                 else (x + abs(f1) if delta > 0 else x - abs(f1)))
            fu = fun(u, **fun_args)

            if fu <= fx:
                if u >= x:
                    min_x = x
                else:
                    max_x = x
                v = w
                w = x
                x = u
                fv = fw
                fw = fx
                fx = fu
            else:
                if u < x:
                    min_x = u
                else:
                    max_x = u

                if (fu <= fw) or (w == x):
                    v = w
                    w = u
                    fv = fw
                    fw = fu
                elif (fu <= fv) or (v == x) or (v == w):
                    v = u
                    fv = fu
            cur_iter += 1

        if cur_iter >= maxiter:
            warnings.warn('Warning: May not have converged. Number of' +
                          'iterations exceeded maximum', Warning)

        final_res = {'Minimum': x,
                     'Final Function Value': fx,
                     'Domain Tolerances': abs(x - min_x),
                     'Iterations': cur_iter,
                     'Method': "Brent"}

    else:

        start = np.array(start).reshape(1, -1)
        num_parms = start.size

        alpha = 1
        beta = 1 + (2 / num_parms)
        gamma = 0.75 - (1 / (2 * num_parms))
        delta = 1 - (1 / num_parms)

        # intialize simplex, function values, and centroid

        h_j = np.where(start != 0, 0.05, 0.00025).reshape(1, -1)
        e_j = np.identity(num_parms)

        simplex = np.concatenate((start, start + (h_j * e_j)),
                                 axis=0)
        sx_w_fval = np.concatenate((simplex, np.empty((num_parms + 1, 1))),
                                   axis=1)
        # sx_w_fval[:, -1] = np.nan
        
        for i in range(num_parms + 1):
            sx_w_fval[i, -1] = fun(simplex[i, :], **fun_args)

        # initial sort
        sx_w_fval = sx_w_fval[sx_w_fval[:, -1].argsort()]

        # when stdev between vertices is small
        domain_conv = False
        # when the function evaluations between vertices is small
        funval_conv = False
        # when the loop reaches a max number of iterations
        cur_iter = 0

        while cur_iter < maxiter and not (funval_conv and domain_conv):
            shrink = False
            centroid = np.mean(sx_w_fval[:-1, :-1], axis=0)
            # Reflection
            x_r = centroid + alpha*(centroid - sx_w_fval[-1, :-1])
            f_x_r = fun(x_r, **fun_args)
            if f_x_r < sx_w_fval[-2, -1] and f_x_r >= sx_w_fval[0, -1]:
                # if reflected is better than worst, replace
                sx_w_fval[-1, :] = np.hstack((x_r, f_x_r))
            elif f_x_r < sx_w_fval[0, -1]:
                # Expansion
                # if reflected point is best yet, try expanding
                x_e = centroid + beta*(x_r - centroid)
                f_x_e = fun(x_e, **fun_args)
                if f_x_e < f_x_r:
                    # if expanded is better, replace with expanded
                    sx_w_fval[-1, :] = np.hstack((x_e, f_x_e))
                else:
                    # replace with reflected
                    sx_w_fval[-1, :] = np.hstack((x_r, f_x_r))
            else:
                # Contraction
                # First, check outside contraction
                # if reflected value is between 2nd worst and worst
                if sx_w_fval[-2, -1] <= f_x_r and f_x_r < sx_w_fval[-1, -1]:
                    x_c_o = centroid + gamma*(x_r - centroid)
                    f_x_c_o = fun(x_c_o, **fun_args)
                    if f_x_c_o <= f_x_r:
                        sx_w_fval[-1, :] = np.hstack((x_c_o, f_x_c_o))
                    else:
                        shrink = True
                else:
                    # Check inside contraction
                    x_c_i = centroid - gamma*(x_r - centroid)
                    f_x_c_i = fun(x_c_i, **fun_args)
                    if f_x_c_i < sx_w_fval[0, -1]:
                        sx_w_fval[-1, :] = np.hstack((x_c_i, f_x_c_i))
                    else:
                        shrink = True

                if shrink:
                    shrunk_vals = (sx_w_fval[0, :-1] +
                                   delta*(sx_w_fval[1:, :-1] -
                                          sx_w_fval[0, :-1]))
                    for i in range(num_parms):
                        sx_w_fval[i + 1, :] = np.hstack((shrunk_vals[i, :],
                                                         fun(shrunk_vals[i, :],
                                                         **fun_args)))

            sx_w_fval = sx_w_fval[sx_w_fval[:, -1].argsort()]

            std_dom = np.std(sx_w_fval[:, :num_parms], axis=0, ddof=1)

            if all(np.array(std_dom)*2 < dom_tol):
                domain_conv = True

            if np.std(sx_w_fval[:, -1], ddof=1)*2 < fun_tol:
                funval_conv = True

            cur_iter += 1

        msg = "Successful"
        if cur_iter >= maxiter:
            msg = 'Warning: May not have converged. Number of' + \
                          'iterations exceeded maximum'
            warnings.warn(msg, Warning)

        final_params = sx_w_fval[0, :]

        final_res = {'Parameters': final_params[:-1],
                     'Final Function Value': final_params[-1],
                     'Domain Tolerances': std_dom,
                     'Function Value Tolerance': np.std(sx_w_fval[:, -1],
                                                        ddof=1),
                     'Iterations': cur_iter,
                     'Message': msg,
                     'Method': "NM"}
    if print_:
        print("Final Results\n", final_res, '\n')

    return final_res


def max_like(start, lf, opt=None, print_=True, like_args=None):
    '''
    Compute maximum likelihood estimation of a function and it's parameters.

    Parameters
    ----------
    start : numpy array
        The starting values for the optimization.
    lf : function
        The likelihood function. An array of the parameters for the function
        must be the first argument.
    opt : dict, optional
        Optional arguments for the optimization. The default is None.
    print_ : bool, optional
        Print the results. The default is True.
    fun_args : dict
        Optional arguments to pass to fun.

    Returns
    -------
    params : list
        A list containing the parameter MLEs, the standard error, and 
        95% CI for the parameters (each row is a parameter).

    '''
    if opt is None:
        opt = {'dom_tol': 1e-8, 'maxiter': 500}
    # initialize variables
    param_res = se = params_ci = hess = None

    # log(start) may be used here aswell because optimization
    # will perform better with log
    
    if like_args is None:
        like_args = {}
    mle_res = AdaptiveNMorBrent(lf, start, opt, print_, fun_args=like_args)
    # note: Hessian needs to be given the parameters
    # in the same scale as the optimization
    # to be able to find the correct derivative
    param_res = 0
    if mle_res['Method'] == "Brent":
        param_res = mle_res["Minimum"]
    else:
        param_res = mle_res["Parameters"]

    try:
        hess = nd.Hessian(lf)(param_res, **like_args)
        np.linalg.inv(hess)
    except np.linalg.LinAlgError as err:
        if 'Singular matrix' in str(err):
            h = False
        else:
            raise

    if np.isnan(hess).any() or not np.isfinite(hess).any():
        h = False
    else:
        if not ((np.linalg.eigvals(hess))).all() > 0:
            h = False
        else:
            h = True

    # se = sqrt(diagonal(covariance matrix)), or,
    # se = sqrt(diagonal(inv(hessian)))
    #
    # ci = estimate +- N(confidence limit,0,1)*se
    #
    # note: when minimizing negative log liklihood in MLE,
    # use the sqrt of the diagonal of the inverse of the Hessian matrix.
    # If maximizing log-likelihood, use negative hessian instead to find
    # standard errors
    #
    # other note: if you perform a log-transformation on the parameters
    # during MLE optimization, you must multiply standard error
    # by exp(paramaters), as well as exponentiate the CI limits.

    num_params = len(np.array(param_res).reshape(-1))
    if h:
        se = np.sqrt(np.diagonal(np.linalg.inv(hess)))
        params_ci = (param_res + np.array([-1, 1]).reshape(2, 1).dot(
                (norm.ppf(1-(1-0.95)/2)*se).reshape(1, num_params))).T

    else:
        se = None
        params_ci = None

    params = [param_res, se, params_ci]

    return params



from __future__ import (division, generators, absolute_import,
                        print_function, with_statement, nested_scopes,
                        unicode_literals)
import datetime
import numpy as np
import os
import pandas as pd
import pickle
import sys

from getpass import getpass
from sqlalchemy import create_engine
from timeit import default_timer as dt


def dsn_getTable(sql, user, dialect, dsn, pw=None, create_engine_args={},
                 read_sql_args={}):
    '''
    Load data into python from a sql database. 
    
    Requires a DSN or connection string. Uses sqlAlchemy to make the
    connection, and closes the connection on completion.
    See https://docs.sqlalchemy.org/en/13/core/engines.html for
    help building connection strings.

    Parameters
    ----------
    sql : str, or list
        A string containing the sql statement to be ran (It should be something
        that returns a table), or a list of sql statements to return multiple
        tables (tables are returned as a list of pandas dataframes, in this
        case).
    user : str
        The username for the database application, if applicable.
    dialect : str
        The database flavor (oracle, mysql, etc.)
    dsn : str, optional
        Can either be a saved DSN from your computer, or a string that
        represents all the relevant information provided in a DSN. 
        It can also be a host:port combination. See sqlalchemy
        documentation for more details.
    pw : str, optional
        The password for the database application that matches the user
        specified, if applicable. If None or '', getpass will ask for a
        password. If there is no password, just leave it blank.
        The default is None.
    create_engine_args : dict
        A dictionary containing arguments to be passed to sqlAlchemy's
        create_engine function.
    read_sql_args : dict
        A dictionary containing arguments to be passed to pandas'
        read_sql function.

    Returns
    -------
    data : pandas dataframe or list
        A dataframe (or list of dataframes) containing the returned data.

    '''
    if pw is None or pw == '':
        pw = getpass(prompt='Password: ')
    
    # TODO: include better support for sqlite, since it has a weird
    # OS-dependent connection string.
    conn = '{dialect}://{user}:{password}@{dsn}'.format(
        dialect=dialect,
        user=user,
        password=pw,
        dsn=dsn
    )

    engine = create_engine(conn, **create_engine_args)

    data = None
    if isinstance(sql, str):
        data = pd.read_sql(sql=sql, con=engine, **read_sql_args)
    else:
        data = []
        for query in sql:
            data.append(pd.read_sql(sql=query, con=engine, **read_sql_args))
    engine.dispose()
    
    return data


def dsn_saveTable(data, tableName, user, dialect, dsn, pw=None,
                  create_engine_args={}, to_sql_args={}):
    '''
    Save pandas dataframe to database.

    Parameters
    ----------
    data : pandas dataframe
        The data to save to a table.
    tableName : str
        The name of the table in the database.
    user : str
        The username for the database application, if applicable.
    dialect : str
        The database flavor (oracle, mysql, etc.)
    dsn : str, optional
        Can either be a saved DSN from your computer, or a string that
        represents all the relevant information provided in a DSN. 
        It can also be a host:port combination. See sqlalchemy
        documentation for more details.
    pw : str, optional
        The password for the database application that matches the user
        specified, if applicable. If None or '', getpass will ask for a
        password. If there is no password, just leave it blank.
        The default is None.
    create_engine_args : dict
        A dictionary containing arguments to be passed to sqlAlchemy's
        create_engine function. Default is {}
    to_sql_args : dict
        A dictionary containing arguments to be passed to pandas'
        read_sql function. Default is {}

    Returns
    -------
    None.

    '''    
    if pw is None or pw == '':
        pw = getpass(prompt='Password: ')
    
    # TODO: include better support for sqlite, since it has a weird
    # OS-dependent connection string.
    conn = '{dialect}://{user}:{password}@{dsn}'.format(
        dialect=dialect,
        user=user,
        password=pw,
        dsn=dsn
    )

    engine = create_engine(conn, **create_engine_args)  
    
    data.to_sql(name=tableName, con=engine, **to_sql_args)
    engine.dispose()
    
    return None


def pretty_print_time(ts, te=None, decimals=0):
    '''
    Convert time in seconds to a clock-like time, i.e. 00:00:00.00 format.

    Parameters
    ----------
    ts : float
        Start time (or elapsed time if te is None).
    te : float, optional
        End time. The default is None.
    decimals : int, optional
        The number of decimal places to use for tracking miliseconds.
        The default is 0

    Returns
    -------
    pretty : str
        Prettified time.

    '''
    if te is None:
        t = ts
    else:
        t = te - ts
    
    hours_placeholder = t/3600.0
    hours = np.floor(hours_placeholder)
    
    minutes_placeholder = (hours_placeholder - hours) * 60.0
    minutes = np.floor(minutes_placeholder)
    
    seconds = np.round((minutes_placeholder - minutes) * 60.0, decimals)
    
    h = str(int(hours))
    if len(h) == 1:
        h = "0" + h
    
    m = str(int(minutes))
    if len(m) == 1:
        m = "0" + m
    
    s = str(int(seconds))
    if len(s) == 1:
        s = "0" + s
    
    if decimals != 0:
        second_decimal = seconds - int(seconds)
        s = '.'.join([s, str(int(second_decimal*10**decimals))])
    
    pretty = h + ":" + m + ":" + s
    return pretty


def checkDir(directory, make=True, verbose=True):
    '''
    Check if a directory exists. If it doesn't, create it.

    Parameters
    ----------
    dir : str
        Directory to check.
    make : bool, optional
        Whether or not to create a missing directory. The default is True.
    verbose : bool
        Print results.

    Returns
    -------
    bool
        True if the directory exists.

    '''
    found = False
    msg = ''
    if not os.path.isdir(directory):
        msg = msg + 'Directory not found. '
        if make:
            os.mkdir(directory)
            msg = msg + 'Directory ' + directory + ' created.'
        else:
            msg = msg + 'Directory ' + directory + 'not created.'
    else:
        found = True
    if verbose:
        print(msg)
    return found

def print_message_with_time(msg, ts, te=None, display_realtime=True,
                            backsn=False, log=False, log_dir="log",
                            log_filename="pmwt.log", log_args="a",
                            time_first=False, decimals=0):
    '''
    Print a message with a timestamp. 

    Parameters
    ----------
    msg : str
        The message you want to print.
    ts : float
        Start time in seconds, or elapsed time if te is None.
    te : float, optional
        End time in seconds. The default is None.
    display_realtime : bool, optional
        Display the system (calendar) time as part of the output.
        The default is True.
    backsn : bool, optional
        Add '\\n'. The default is False.
    log : bool, optional
        Save message to a log file. The default is False.
    log_dir : str, optional
        The directory for the log file. The default is "log".
    log_filename : str, optional
        The name of the logfile. The default is "pmwt.log".
    log_args : str, optional
        The read/write specification, for example, wb, w, a, etc.
        The default is "a" for append. See open function in python for more
        details.
    time_first : bool, optional
        Place time at the begininng of the message. The default is False.
    decimals : int, optional
        The number of decimal places to use for tracking miliseconds.
        The default is 0

    Returns
    -------
    printed : str
        Printed message with time.

    '''
    date_time = datetime.datetime.now().strftime("%I:%M:%S %p (%b %d)")
    if display_realtime:
        time_str = pretty_print_time(ts, te, decimals) + " Current Time: " + date_time
    else:
        time_str = pretty_print_time(ts, te, decimals)
    if time_first:
        printed = time_str + msg
    else:
        printed = msg + time_str
    if backsn:
        printed = printed + "\n"
    sys.stdout.write(printed)
    sys.stdout.flush()
    checkDir(log_dir)
    logfile = log_dir + "/" + log_filename
    logfile = logfile.replace("//", "/")
    if log:
        with open(logfile, log_args) as f:
            f.write(printed)
    return printed
    

def print_time(*args, **kwargs):
    '''
    Wrapper for print_message_with_time. Shortened for ease of use.

    Parameters
    ----------
    *args : 
        Positional arguments (passed to print_message_with_time).
    **kwargs : 
        Keyword Arguments (passed to print_message_with_time).

    Returns
    -------
    None.

    '''
    print_message_with_time(*args, **kwargs)


def save_checkpoint(obj, msg, ts=None, te=None, decimals=0,
                    display_realtime=True, backsn=False,
                    log=True, log_dir="log", log_filename="", log_args="a",
                    time_first=False, checkpoint_dir="checkpoints",
                    checkpoint_filename="chkpnt", checkpoint_extension="pkl",
                    checkpoint_args="wb"):
    '''
    Save a python object to a pickle file as a checkpoint. Log message as 
    logfile with the export.

    Parameters
    ----------
    obj : any python object (must be pickleable)
        The object to pickle.
    msg : str
        The message you want to print.
    ts : float
        Start time in seconds, or elapsed time if te is None.
    te : float, optional
        End time in seconds. The default is None.
    decimals : int, optional
        The number of decimal places to use for tracking miliseconds.
        The default is 0
    display_realtime : bool, optional
        Display the system (calendar) time as part of the output.
        The default is True.
    backsn : bool, optional
        Add '\\n'. The default is False.
    log : bool, optional
        Save message to a log file. The default is False.
    log_dir : str, optional
        The directory for the log file. The default is "log".
    log_filename : str, optional
        The name of the logfile. The default is "pmwt.log".
    log_args : str, optional
        The read/write specification, for example, wb, w, a, etc.
        The default is "a" for append. See open function in python for more
        details.
    time_first : bool, optional
        Place time at the begininng of the message. The default is False.
    checkpoint_dir : str, optional
        The directory to save the checkpoint. The default is "checkpoints".
    checkpoint_filename : str, optional
        The filename for the checkpoint. The default is "chkpnt".
    checkpoint_extension : str, optional
        The file extension for the checkpoint. The default is "pkl".
    checkpoint_args : str, optional
        The read/write specification, for example, wb, w, a, etc.
        See open function in python for more details.
        The default is "wb" for write binary.

    Returns
    -------
    None.

    '''
    print_message_with_time(msg=msg, ts=ts, te=te, decimals=decimals,
                            display_realtime=display_realtime,
                            backsn=backsn, log=log, log_dir=log_dir,
                            log_filename=log_filename,
                            log_args=log_args)
    checkDir(checkpoint_dir)
    file = checkpoint_dir + '/' + checkpoint_filename + "." + checkpoint_extension
    file = file.replace('//', '/')
    if checkpoint_extension == 'pkl':
        with open(file, checkpoint_args) as f:
            pickle.dump(obj, f, protocol=pickle.HIGHEST_PROTOCOL)


def updateProgBar(curIter, totalIter, t0, barLength=20, decimals=0):
    '''
    Update progress bar. Place this function anywhere in a loop where you want
    to keep track of the loop's progress.

    Parameters
    ----------
    curIter : int
        The current iteration.
    totalIter : int
        The total number of iterations. 
    t0 : numeric
        The start time of the operation (in seconds).
    barLength : int, optional
        The length of the progress bar. The default is 20.
    decimals : int, optional
        The number of decimal places to use for tracking miliseconds.
        The default is 0

    Returns
    -------
    None.

    '''
    status = "Working..."
    progress = float(curIter)/float(totalIter)
    if isinstance(progress, int):
        progress = float(progress)
    if progress >= 1:
        progress = 1
        status = "Finished!..."
    block = int(round(barLength*progress))
    text = "\rPercent: [{0}] {1:.2f}% iter: {2}/{3} {4} Elapsed: {5}, Estimated: {6}".format(
        "#"*block + "-"*(barLength - block), 
        round(progress*100.0, 2), curIter, totalIter, status,
        pretty_print_time(t0, dt(), decimals=decimals),
        pretty_print_time((dt()-t0)/curIter * (totalIter - curIter),
                          decimals=decimals))
    if progress >= 1:
        sys.stdout.write(text + "\r\n")
        sys.stdout.flush()
    else:
        sys.stdout.write(text)
        sys.stdout.flush()
        
        
def move_columns_to_end(data, col):
    '''
    Rearrange a pandas dataframe by moving a column (or list of columns)
    to the end of the dataframe.

    Parameters
    ----------
    data : pandas dataframe
        Data to rearrange.
    col : str or list
        String containing one column to move, or list containing several.

    Returns
    -------
    data : pandas dataframe
        Data with rearranged columns.

    '''
    if not isinstance(col, list):
        col = list(col)
    data = data[[c for c in data if c not in col] + col]
    return data


# Searches a list of text arguments and returns a list that includes/excludes
# the list elements that contain the keywords
def keywordSearch(text_list, include_keywords, exclude_keywords=None):
    '''
    Searches a list of text arguments and returns a list that includes/excludes
    the list elements that contain the keywords.
    
    Useful when searching for column names in a large dataframe, for example.

    Parameters
    ----------
    text_list : list
        List of string values.
    include_keywords : list
        List of keywords to search for.
    exclude_keywords : list, optional
        List of keywords that might be similar to include_keywords, but should
        be excluded nonetheless. The default is None.

    Returns
    -------
    found_list : list
        Returns items from the original list that match the keywords.

    '''
    found_list = []
    for text in text_list:
        is_in_i_keys = sum((1 if i_key in text else 0 for i_key in include_keywords))
        is_in_e_keys = 0
        if exclude_keywords is not None:
            is_in_e_keys = sum((1 if e_key in text else 0 for e_key in exclude_keywords))
        if is_in_i_keys > 0 and is_in_e_keys == 0:
            found_list.append(text)
    return found_list


def toFrame(todf, fromdf):
    '''
    Convert numpy array to pandas dataframe using past dataframe structure. 

    Useful when you want to store the dataframe values as a seperate numpy
    array, and make adjustments to the values, but want to convert back to a
    dataframe once the calculations are completed. 

    Parameters
    ----------
    todf : numpy array
        Data to convert to pandas dataframe.
    fromdf : pandas dataframe
        Dataframe containing structure to convert back to.

    Returns
    -------
    pandas dataframe
        The converted dataframe.

    '''
    return pd.DataFrame(todf, index=fromdf.index, columns=fromdf.columns)


def inferFeatureType(X, n_unique=None):
    '''
    Infer feature type of a column based on it's `dytpe`. The inferred
    datatypes are `categorical`, `numeric`, or `date`.

    
    Parameters
    ----------
    X : dataframe or ndarray
        The dataframe containing columns whose datatype needs to be inferred.
    n_unique : int
        The number of unique values to use as a cutoff for numeric columns,
        for example, if the number of unique values for a numeric column is
        greater than n_unique, consider it numeric. Default is None.
    
    Returns:
    --------
    d_type : str
        String representation of the d_type
    
    numpy arrays.dtypes reference table:
    
    ? : boolean
    b : signed byte (sometimes interchangeable with boolean, it seems)
    B : unsigned byte
    i : signed integer
    u : unsigned integer
    f : float
    c : complex float
    m : timedelta
    M : datetime
    O : object
    S : zero-terminated bytes
    a : zero-terminated bytes
    U : unicode string
    V : raw data (void)
    
    '''
    d_type = None
    if X.dtype.kind in "OSUVaBb": # ?/b numeric or categorical?
        d_type = "categorical"
    elif X.dtype.kind in "mM":
        d_type = "date"
    else:
        if n_unique is None:
            d_type = "numeric"
        else:
            n_uni_obs = len(pd.unique(X))
            if n_uni_obs <= n_unique:
                d_type = "categorical"
            else:
                d_type = "numeric"
    return d_type


def reshape_to_vect(ar, axis=1):
    '''
    Flatten or reshape an array to be a vector (or transpose an already
    flat array).

    Parameters
    ----------
    ar : numpy array
        An array to be reshaped.
    axis : int, optional
        The direction to flatten. The default is 1.

    Returns
    -------
    ar : numpy array
        The flattened array.

    '''
    # TODO: Need to look at the use case for this again. There may be a more
    # efficient way to do this.
    if len(ar.shape) == 1:
        if axis == 1:
            return ar.reshape(ar.shape[0], 1)
        elif axis == 0:
            return ar.reshape(1, ar.shape[0])
        else:
            raise('Invalid axis dimension, either 0 or 1')
    return ar


def decision_boundary_1D(x, thres=0.5):
    '''
    Divides probabilities `x` into two classes based on threshold value. Everything
    above the threshold is the positive class.
    '''
    y = np.zeros(shape=(len(x), 1))
    y[x >= thres] = 1
    return y


def decision_boundary(x):
    '''
    Divides probabilities `x` into classes based on the max probability value.
    '''
    y = np.zeros(shape=x.shape)
    y[np.arange(len(x)), x.argmax(1)] = 1
    return y


def one_hot_encode(x):
    '''
    The first column is the 0th class, 2nd column is the 1st class, etc...
    '''
    n = np.max(x) + 1
    return np.eye(n)[x]


def one_hot_decode(x):
    '''
    The first column is the 0th class, 2nd column is the 1st class, etc...
    '''
    return np.argmax(x, axis=1)



import numpy as np


def gradient_descent(wb, dwdb, optimizer_args):
    '''
    The simple gradient descent update.

    Parameters
    ----------
    wb : dict
        A dictionary of the weights/biases for each layer.
    dwdb : dict
        A dictionary of the gradients with respect to the weights and
        biases.
    optimizer_args : dict
        Optional optimizer configurations.

    Returns
    -------
    wb : dict
        A dictionary of the updated weights and biases.

    '''
    for key in wb.keys():
        wb[key] = wb[key] - optimizer_args['learning_rate']*dwdb[key]
    return wb


def adam(wb, dwdb, optimizer_args):
    '''
    The Adam optimizer. (Adaptive Moment Estimation)

    Parameters
    ----------
    wb : dict
        A dictionary of the weights/biases for each layer.
    dwdb : dict
        A dictionary of the gradients with respect to the weights and
        biases.
    optimizer_args : dict
        Optional optimizer configurations.

    Returns
    -------
    wb : dict
        A dictionary of the updated weights and biases.

    '''
    for key in wb.keys():
        optimizer_args['mt'][key] = (optimizer_args['beta1']*optimizer_args['mt'][key] +
                                     (1-optimizer_args['beta1'])*dwdb[key])
        optimizer_args['vt'][key] = (optimizer_args['beta2']*optimizer_args['vt'][key] +
                                     (1-optimizer_args['beta2'])*(dwdb[key]**2))
        mhat = optimizer_args['mt'][key]/(1-(optimizer_args['beta1']**(optimizer_args['i']+1)))
        vhat = optimizer_args['vt'][key]/(1-(optimizer_args['beta2']**(optimizer_args['i']+1)))
        wb[key] = wb[key] - optimizer_args['learning_rate']*(mhat/(np.sqrt(vhat)+optimizer_args['eps']))
    return wb


"""
Statistical Distribution Functions
"""

from __future__ import (division, generators, absolute_import,
                        print_function, with_statement, nested_scopes,
                        unicode_literals)
import numpy as np

from scipy.stats import norm as normal, lognorm, gamma
from scipy.special import xlogy, loggamma

from .tmath import norm


# #########################################
# ########## General Functions ############
# #########################################


# This is a faster approximation of the normal quantile function.
# Accurate to 2 or 3 digits. For exact solutions, use dwrap or scipy
def qnorm_aprox(p, mu=0, sigma=1, lt=True):
    '''
    This is a faster approximation of the normal quantile function (At least
    this was true at the time I last benchmarked it). Accurate to 2 or 3
    digits. For exact solutions, use dwrap or scipy.

    Parameters
    ----------
    p : float or array-like
        percentile (or probability) of interest.
    mu : float, optional
        Mean of the normal distribution. The default is 0.
    sigma : float, optional
        Standard deviation of the normal distribution. The default is 1.
    lt : bool, optional
        Lower tail of the distribution. The default is True.

    Returns
    -------
    quant : float or numpy array
        Return quantile of interest.

    '''
    p = np.array(p)
    if not lt:
        p = 1 - p
    a = 0.147
    x = 2.0*p - 1.0
    b = (2.0/(np.pi*a)) + (np.log(1.0 - x**2)/2.0)
    c = (np.log(1.0 - x**2)/a)
    i_erf = np.sign(x)*np.sqrt(np.sqrt(b**2 - c) - b)
    quant = mu + sigma*np.sqrt(2)*i_erf
    return quant


def pnorm_approx(q, mu=0, sigma=1, lt=True, log=False):
    '''
    This is a fast approximation of the normal cdf, accurate to 1.5e-7.
    For exact solutions, use dwrap or SciPy.

    Parameters
    ----------
    q : float
        Quantile of interest.
    mu : float, optional
        Mean of the normal distribution. The default is 0.
    sigma : float, optional
        Standard deviation of the normal distribution. The default is 1.
    lt : bool, optional
        Use lower tail. The default is True.
    log : bool, optional
        Return log of the value. The default is False.

    Returns
    -------
    prob : float
        The probability at the given quantile.

    '''
    q = np.array(q)
    p = 0.3275911
    a1 = 0.254829592
    a2 = -0.284496736
    a3 = 1.421413741
    a4 = -1.453152027
    a5 = 1.061405429
    x = (q - mu)/(sigma * np.sqrt(2))
    sx = np.sign(x)
    ax = np.abs(x)
    t = (1 / (1 + p*ax))
    erf = sx * (1 - (a1*t +
                a2*(t**2) +
                a3*(t**3) +
                a4*(t**4) +
                a5*(t**5)) * np.exp(-(ax**2)))

    prob = 0.5*(1.0 + erf)
    if not lt:
        prob = 1.0 - prob
    if log:
        prob = np.log(prob)
    return prob


def dpoibin_PA(k, p):
    '''
    Poisson approximation of the Poisson-Binomial probability mass function.
    Fast, but possibly inaccurate.

    Parameters
    ----------
    k : numpy array or int
        The poisson counts.
    p : numpy array or float
        Binomial probabilities.

    Returns
    -------
    prob : float
        The mass of the poisson-binomial distribution.

    '''
    lmbda = np.sum(p)
    prob = np.exp(k * np.log(lmbda) - lmbda - loggamma(k + 1.0))
    return prob


def ppoibin_RNA(k, p, p_weight=None):
    '''
    Refined normal approximation of Poisson-Binomial Cumulative Distribution.
    This code is adapted from the R package 'poibin'.
    Very fast for p > 10000.

    Parameters
    ----------
    k : numpy array or int
        Poisson counts.
    p : numpy array or float
        Binomial probabilities.
    p_weight : numpy array, optional
        Weights for the probabilities. Same length as p. The default is None.

    Returns
    -------
    vkk_r : numpy array or float
        cumulative probabilities.

    '''
    p = np.array(p).reshape(-1, )
    k = np.array(k).reshape(-1, )
    if any(p < 0) or any(p > 1):
        raise("Invalid Probabilities (0 <= P(x) <= 1)")

    if p_weight is None:
        p_weight = np.repeat(1, len(p))

    pp = np.repeat(p, p_weight)
    muk = np.sum(pp)
    sigmak = np.sqrt(np.sum(pp * (1 - pp)))
    gammak = np.sum(pp * (1 - pp) * (1 - 2 * pp))
    kk1 = (k + 0.5 - muk)/sigmak
    vkk_r = (pnorm_approx(kk1) + gammak/(6 * sigmak**3) * (1 - kk1**2) *
             dwrap(kk1, [0, 1], "pdf", "normal"))
    vkk_r[np.where(vkk_r < 0)] = 0
    vkk_r[np.where(vkk_r > 1)] = 1
    return vkk_r


def dpoibin_FT(k, p):
    '''
    The probability mass function for the poisson-binomial distribution.
    This function is written by the author of this package, and uses the
    discrete fourier transform.

    Parameters
    ----------
    k : numpy array or int
        Poisson counts.
    p : numpy array or float
        Binomial probabilities.

    Returns
    -------
    karray : float or numpy array
        The mass of the poisson-binomial distribution.

    '''
    n = len(p)
    nk = len(k)
    inp1 = (1/(n + 1))
    L = np.arange(0, n + 1, 1)
    C = np.exp((2*1j*np.pi)/(n + 1))
    karray = np.zeros(nk)
    k_bool = np.array((k < 0) | (k > n))
    for i in range(nk):
        if k_bool[i]:
            karray[i] = 0
        else:
            prodp = np.array([np.prod(1 + (C**x - 1)*p) for x in L])
            c_calc = C**(-L*k[i])
            sumc = np.sum(c_calc * prodp)
            res = inp1*sumc
            karray[i] = res.real
    if any(karray < 0):
        karray[np.where(karray < 0)] = 0
    return karray


# This function is adapted from Mark Lodato of StackExchange, Nov 7'16
# It's fast, but runs slower than 1 second once p > 25000
# Link: https://stats.stackexchange.com/questions/242233/efficiently-computing-poisson-binomial-sum
# Computes both the pdf and the cdf
def dpoibin_exact(k, p, cdf=False):
    '''
    The probability mass function for the poisson-binomial distribution.
    
    This function is adapted from Mark Lodato of StackExchange, Nov 7'16
    It's fast, but runs slower than 1 second once len(p) > 25000
    Link: https://stats.stackexchange.com/questions/242233/efficiently-computing-poisson-binomial-sum
    
    Adapted to compute both the pdf and the cdf

    Parameters
    ----------
    k : numpy array or int
        Poisson counts.
    p : numpy array or float
        Binomial probabilities.
    cdf : bool, optional
        Compute the cdf. Default is False.

    Returns
    -------
    karray : float or numpy array
        The mass of the poisson-binomial distribution.

    '''
    k = np.array(k).reshape(-1,)
    p = np.array(p).reshape(-1,)
    
    n = p.shape[0]
    m = k.shape[0]
    karray = np.zeros(m)
    pmf = np.zeros(n + 1, dtype=float)
    pmf[0] = 1.0
    for i in range(n):
        success = pmf[:i + 1] * p[i]
        pmf[:i + 2] *= (1 - p[i])  
        pmf[1:i + 2] += success
    for i, k_elem in enumerate(k):
        if cdf:
            karray[i] = pmf[:int(k_elem+1)].sum()
            if any(k > n):
                karray[np.where(k > n)] = 1
        else:
            if k_elem <= n:
                karray[i] = pmf[k_elem]
            if any(k > n):
                karray[np.where(k > n)] = 0
    if any((karray < 0) | (k < 0)):
        karray[np.where((karray < 0) | (k < 0))] = 0
    return karray


def dwrap(data, params, disttype, funct, log=False):
    '''
    This function is meant to be similar to the R distribution functions,
    such as dnorm, pnorm, qnorm, etc. It calculates variations of the 
    cdf or pdf depending on the funct selected.
    
    I have found that writing the distributions in plain math is sometimes 
    faster in python than using the scipy implementations, which is why
    this function originally existed. I wrote my own versions of the
    distributions, except where the distributions were complicated and it
    wasn't worth it at the time (and then this function uses scipy). However,
    as time goes on, that will probably change. Also, the scipy.stats
    implementations and documentations are quite complete, so unless you're
    feeling adventurous, it's probably a good idea to just use scipy.

    Parameters
    ----------
    data : numpy array or pandas dataframe/series
        Numeric values representing the data of interest, either a random
        variable, probability, or quantile.
    params : numpy array or pandas dataframe/series
        the parameters of the function of interest (shape, scale, etc.)
        - weibull: (shape, scale)
        - 
    disttype : str
        the distribution type, which currently includes pdf, cdf, inv-cdf, sf,
        left-truncated-cdf, left-truncated-inv-cdf. (Note: not all of these
        options may be available for all funct options)
    funct : str
        the distribution function, which currently includes weibull, 
        exponential, log-normal (as lnorm), normal, and gamma
    log : bool, optional
        Whether to use the log of the distribution or not. The default is
        False.

    Raises
    ------
    ValueError
        Raised when invalid distribution type or function is chosen.

    Returns
    -------
    numpy array 
        an array containing the evaluation of the distribution.

    '''
    try:
        num_parms = params.shape[1]
        params = np.array(params).reshape(-1, num_parms)
    except (IndexError, AttributeError):
        params = np.array(params).reshape(1, -1)

    if funct == "weibull":
        shape = params[:, 0]
        scale = params[:, 1]
        if disttype == "pdf":
            if log:
                return (np.log(shape) - np.log(scale) +
                        xlogy(shape - 1.0, data/scale) -
                        (data/scale)**shape)
            else:
                return ((shape/scale)*((data/scale)**(shape - 1.0)) *
                        np.exp(-(data/scale)**shape))
        elif disttype == "cdf":
            if log:
                return np.log((1.0 - np.exp(-(data/scale)**shape)))
            else:
                return (1.0 - np.exp(-(data/scale)**shape))
        elif disttype == "sf":
            if log:
                return -(data/scale)**shape
            else:
                return np.exp(-(data/scale)**shape)
        elif disttype == "inv-cdf":
            if log:
                return np.log(scale) + ((1/shape)*np.log(-np.log(1 - data)))
            else:
                return scale*(-np.log(1 - data))**(1/shape)
        # The left-truncted distribution is useful for evaluating the
        # censored items (probability of failing given it lived this long)
        elif disttype == "left-truncated-cdf":
            a = params[:, 2]
            if log:
                return np.log(1 - (np.exp(((a/scale)**shape) - ((data/scale)**shape))))
            else:
                return 1 - (np.exp(((a/scale)**shape) - ((data/scale)**shape)))
        elif disttype == "left-truncated-inv-cdf":
            a = params[:, 2]
            if log:
                return np.log(scale) + (1/shape)*np.log((((a/scale)**shape) - np.log(1 - data)))
            else:
                return scale*(((a/scale)**shape) - np.log(1 - data))**(1/shape)
        else:
            raise ValueError("Not a valid distribution type")
    elif funct == "exponential":
        rate = params[:, 0]

        if disttype == "pdf":
            if log:
                return np.log(rate) - rate*data
            else:
                return rate * np.exp(-rate*data)
        elif disttype == "cdf":
            if log:
                return np.log(1.0 - np.exp(-rate*data))
            else:
                return 1.0 - np.exp(-rate*data)
        elif disttype == "sf":
            if log:
                return -rate*data
            else:
                return np.exp(-rate*data)
        else:
            raise ValueError("Not a valid distribution type")
    elif funct == "lnorm":
        mu = params[:, 0]
        sigma = params[:, 1]
        if disttype == "pdf":
            if log:
                return (-(np.log(data) + np.log(sigma) +
                          0.5*np.log(2*np.pi)) -
                        (((np.log(data) - mu)**2)/(2*sigma**2)))
            else:
                return ((1/(data*sigma*np.sqrt(2*np.pi))) *
                        np.exp(-((np.log(data) - mu)**2)/(2*sigma**2)))
        elif disttype == "cdf":
            if log:
                return lognorm.logcdf(x=data, scale=np.exp(mu), s=sigma)
            else:
                return lognorm.cdf(x=data, scale=np.exp(mu), s=sigma)
        elif disttype == "sf":
            if log:
                return lognorm.logsf(x=data, scale=np.exp(mu), s=sigma)
            else:
                return lognorm.sf(x=data, scale=np.exp(mu), s=sigma)
        else:
            raise ValueError("Not a valid distribution type")
    elif funct == "normal":
        mu = params[:, 0]
        sigma = params[:, 1]
        if disttype == "pdf":
            if log:
                return (-0.5*np.log(2*np.pi) - np.log(sigma) -
                        0.5*((data - mu) / sigma)**2)
            else:
                return ((1/(np.sqrt(2*np.pi)*sigma)) *
                        (np.exp(-((data - mu)**2)/(2*(sigma**2)))))
        elif disttype == "cdf":
            if log:
                return normal.logcdf(x=data, loc=mu, scale=sigma)
            else:
                return normal.cdf(x=data, loc=mu, scale=sigma)
        elif disttype == "sf":
            if log:
                return normal.logsf(x=data, loc=mu, scale=sigma)
            else:
                return normal.sf(x=data, loc=mu, scale=sigma)
        elif disttype == "inv-cdf":
            if log:
                return normal.logppf(q=data, loc=mu, scale=sigma)
            else:
                return normal.ppf(q=data, loc=mu, scale=sigma)
        else:
            raise ValueError("Not a valid distribution type")
    elif funct == "gamma":
        param1 = params[:, 0]
        param2 = params[:, 1]
        if disttype == "pdf":
            if log:
                return gamma.logpdf(x=data, a=param1, scale=param2)
            else:
                return gamma.pdf(x=data, a=param1, scale=param2)
        elif disttype == "cdf":
            if log:
                return gamma.logcdf(x=data, a=param1, scale=param2)
            else:
                return gamma.cdf(x=data, a=param1, scale=param2)
        elif disttype == "sf":
            if log:
                return gamma.logsf(x=data, a=param1, scale=param2)
            else:
                return gamma.sf(x=data, a=param1, scale=param2)
        else:
            raise ValueError("Not a valid distribution type")
    else:
        raise ValueError("Not a valid distribution")


# #########################################
# ######## Likelihood Functions ###########
# #########################################


def likelihood_bernoulli(y_true, y_score, neg=True, log=True):
    '''
    The likelihood for a binary output or bernoulli model 
    
    Parameters
    ----------
    y_true : numpy array (numeric)
        The true values (the binary observations).
    y_score : numpy array or pandas dataframe
        The predicted probabilities.
    neg : bool, optional
        Return negative likelihood. The default is True.
    log : bool, optional
        Return log-likelihood. The default is True.

    Returns
    -------
    float
        The likelihood.

    '''
    
    ### Alternate formulation (placing here for my notes)
    # loglike = y_true*y_score - np.log(1 + np.exp(y_score))
    loglike = np.sum(xlogy(y_true, y_score) + xlogy(1.0 - y_true, 1.0 - y_score))

    if not log:
        loglike = np.exp(loglike)

    if neg:
        return -loglike
    else:
        return loglike


def glm_likelihood_bernoulli(parms, X, Y, lamb=1, l_p=1, neg=True, log=True):
    '''
    The likelihood for a logistic regression or bernoulli model with a penalty
    term (can accept any norm, default is 1 for L1)
    
    Parameters
    ----------
    parms : numpy array (numeric)
        The coefficients (including intercept, which is first)
    X : numpy array (numeric)
        The independent variables (or feature matrix), where the first column
        is a dummy column of 1's (for the intercept).
    Y : numpy array or pandas dataframe
        The response value (should be 0 or 1, but could be float as well if 
        you're willing to deal with those consequences).
    lamb : int, optional
        The size of the penalty (lambda). Note this is the inverse of the
        common sklearn parameter C (i.e. C=1/lambda. The default is 1.
    l_p : int, optional
        The mathmatical norm to be applied to the coefficients.
        The default is 1, representing an L1 penalty.
    neg : bool, optional
        Return negative likelihood. The default is True.
    log : bool, optional
        Return log-likelihood. The default is True.

    Returns
    -------
    float
        The likelihood.

    '''
    #intercept = parms[0]
    betas = parms[1:]
    
    mu = X.dot(parms)
    Ypred = 1.0/(1.0 + np.exp(-mu))
    ### Alternate formulation (placing here for my notes)
    # loglike = Y*mu - np.log(1 + np.exp(mu))
    loglike = np.sum(xlogy(Y, Ypred) + xlogy(1.0 - Y, 1.0 - Ypred)) - lamb*norm(betas, l_p)

    if not log:
        loglike = np.exp(loglike)

    if neg:
        return -loglike
    else:
        return loglike


def likelihood_poisson(y_true, y_score, neg=True, log=True):
    '''
    The likelihood for a count output or poisson model 
    
    Parameters
    ----------
    y_true : numpy array (numeric)
        The true values (the poisson observations).
    y_score : numpy array or pandas dataframe
        The estimated value/values (lambda).
    neg : bool, optional
        Return negative likelihood. The default is True.
    log : bool, optional
        Return log-likelihood. The default is True.

    Returns
    -------
    float
        The likelihood.

    '''
    loglike = np.sum(y_true*np.log(y_score) - y_score - loggamma(y_true + 1))

    if not log:
        loglike = np.exp(loglike)

    if neg:
        return -loglike
    else:
        return loglike


def glm_likelihood_poisson(parms, X, Y, lamb=1, l_p=1, neg=True, log=True):
    '''
    The likelihood for a poisson regression model with a penalty
    term (can accept any norm, default is 1 for L1)
    
    Parameters
    ----------
    parms : numpy array (numeric)
        The coefficients (including intercept, which is first)
    X : numpy array (numeric)
        The independent variables (or feature matrix), where the first column
        is a dummy column of 1's (for the intercept).
    Y : numpy array or pandas dataframe
        The response value (should be 0 or 1, but could be float as well if 
        you're willing to deal with those consequences).
    lamb : int, optional
        The size of the penalty (lambda). Note this is the inverse of the
        common sklearn parameter C (i.e. C=1/lambda. The default is 1.
    l_p : int, optional
        The mathmatical norm to be applied to the coefficients.
        The default is 1, representing an L1 penalty.
    neg : bool, optional
        Return negative likelihood. The default is True.
    log : bool, optional
        Return log-likelihood. The default is True.

    Returns
    -------
    float
        The likelihood.

    '''
    #intercept = parms[0]
    betas = parms[1:]
    
    mu = X.dot(parms)
    
    loglike = np.sum(Y*mu - np.exp(mu) - loggamma(Y + 1)) - lamb*norm(betas, l_p)

    if not log:
        loglike = np.exp(loglike)

    if neg:
        return -loglike
    else:
        return loglike


def likelihood_gaussian(y_true, y_score, neg=True, log=True, sigma=None):
    '''
    The likelihood for a normal(ish) output or gaussian model 
    
    Parameters
    ----------
    y_true : numpy array (numeric)
        The true values (the normal observations).
    y_score : numpy array or pandas dataframe
        The estimated value/values (mean of the normal distribution).
    neg : bool, optional
        Return negative likelihood. The default is True.
    log : bool, optional
        Return log-likelihood. The default is True.

    Returns
    -------
    float
        The likelihood.

    '''
    if sigma is None:
        # The MLE estimate of sigma
        sigma = np.sqrt(np.mean((y_true - y_score)**2))

    loglike = np.sum((-0.5*np.log(2*np.pi) - np.log(sigma) - 0.5*((y_true - y_score) / sigma)**2))

    if not log:
        loglike = np.exp(loglike)

    if neg:
        return -loglike
    else:
        return loglike


def glm_likelihood_gaussian(parms, X, Y, lamb=1, l_p=1, sigma=None, neg=True,
                            log=True):
    '''
    The likelihood for a gaussian regression model with a penalty
    term (can accept any norm, default is 1 for L1)
    
    Parameters
    ----------
    parms : numpy array (numeric)
        The coefficients (including intercept, which is first)
    X : numpy array (numeric)
        The independent variables (or feature matrix), where the first column
        is a dummy column of 1's (for the intercept).
    Y : numpy array or pandas dataframe
        The response value (should be 0 or 1, but could be float as well if 
        you're willing to deal with those consequences).
    lamb : int, optional
        The size of the penalty (lambda). Note this is the inverse of the
        common sklearn parameter C (i.e. C=1/lambda. The default is 1.
    l_p : int, optional
        The mathmatical norm to be applied to the coefficients.
        The default is 1, representing an L1 penalty.
    neg : bool, optional
        Return negative likelihood. The default is True.
    log : bool, optional
        Return log-likelihood. The default is True.

    Returns
    -------
    float
        The likelihood.

    '''
    #intercept = parms[0]
    betas = parms[1:]
    
    mu = X.dot(parms)
    if sigma is None:
        # The MLE estimate of sigma
        sigma = np.sqrt(np.sum((Y - mu)**2)/(Y.shape[0] - len(betas) - 1))
    
    loglike = np.sum((-0.5*np.log(2*np.pi) - np.log(sigma) - 0.5*((Y - mu) / sigma)**2)) - lamb*norm(betas, l_p)
    
    if not log:
        loglike = np.exp(loglike)

    if neg:
        return -loglike
    else:
        return loglike
    

def ExactMLE_exp(data, censored):
    '''
    Maximum Likelihood Estimate of an exponential distribution (with an option
    for right-censoring).

    Parameters
    ----------
    data : numpy array
        The data for the exponential distribution (usually time between events,
        or similar).
    censored : numpy array
        An array of boolean values where 1 means the event occured and 0 means
        the event is censored (right).

    Returns
    -------
    list
        A list containing (in order) the MLE, the standard error, and the 
        95% confidence interval for the estimate.

    '''
    f = data[censored == 1]
    c = data[censored == 0]
    nf = f.size
    rate = nf/(np.sum(f) + np.sum(c))
    se = rate / np.sqrt(nf)
    params_ci = (rate + np.array([-1, 1])*norm.ppf(1-(1-0.95)/2)*se)
    return [rate, se, params_ci]


# #########################################
# ########## Posterior Functions ##########
# #########################################


def posterior_logreg_lasso(parms, X, Y, l_scale=0.5, neg=False, log=True):
    '''
    The posterior density for a logistic regression model with an L1 penalty
    term.
    
    Parameters
    ----------
    parms : numpy array (numeric)
        The coefficients (including intercept, which is first)
    X : numpy array (numeric)
        The independent variables (or feature matrix), where the first column
        is a dummy column of 1's (for the intercept).
    Y : numpy array or pandas dataframe
        The response value (should be 0 or 1, but could be float as well if 
        you're willing to deal with those consequences).
    l_scale : float, optional
        The value of the scale parameter in the Laplace distribution.
        A common choice for the laplace prior is scale = 2/lambda, 
        because it can be shown that this is the MAP estimate where
        a laplace prior is equivalent to performing lasso regression.
        lambda is the L1 penalty, or scale = 2*C (where C is the penalty term
        in sklearn).
        
        I find that when scale == C, you get more similar results to
        the model output in sklearn, and based on
        my research into their implementation, I think it is because the
        sklearn implementation actually scales lambda by 1/2 (see the user
        guide here for more details:
        https://scikit-learn.org/stable/modules/linear_model.html),
        but i'm not 100% sure on this. This parameterization is similar to
        scale = stddev/lambda or scale = stddev*C, where they set stddev to
        1 instead of 2. The default value is 1.
    neg : bool, optional
        Return negative density. The default value is False.
    log : bool, optional
        Return log-density. The default value is True.

    Returns
    -------
    float
        The posterior density (height of the posterior density)

    '''
    n_mu = 0
    n_sigma = 10
    l_loc = 0
    
    intercept = parms[0]
    betas = parms[1:]
    
    mu = X.dot(parms)
    Ypred = 1.0/(1.0 + np.exp(-mu))
    like = np.sum(Y*np.log(Ypred) + (1.0 - Y)*np.log(1.0 - Ypred))
    # normal prior on the intercept
    int_prior = (-0.5*np.log(2.0*np.pi) - np.log(n_sigma) - 0.5*((intercept - n_mu) / n_sigma)**2)
    # Laplace prior 
    parm_prior = np.sum(-np.log(2*l_scale) - (np.abs(betas - l_loc)/l_scale))
    post = like + parm_prior + int_prior 
    
    if not log:
        post = np.exp(post)

    if neg:
        return -post
    else:
        return post


def ap_logreg_lasso(parms, X, Y, l_scale=None, neg=False, log=True):
    '''
    The posterior density for a logistic regression model with an L1 penalty
    term. However, unlike the posterior_logreg_lasso function, this is made to
    addaptively learn the optimal L1 penalty. Therefore, the L1 penalty is
    in the parms variable, at the end of the array
    
    Parameters
    ----------
    parms : numpy array (numeric)
        The model coefficients (including intercept, which is first, and the
        scale of the laplace distribution, which is last)
    X : numpy array (numeric)
        The independent variables (or feature matrix), where the first column
        is a dummy column of 1's (for the intercept).
    Y : numpy array or pandas dataframe
        The response value (should be 0 or 1, but could be float as well if 
        you're willing to deal with those consequences).
    l_scale : float, optional
        ---THIS IS NOT USED. ONLY HERE FOR CONVENIENCE OF THE USER WHEN
        EXPERIMENTING BETWEEN THE ADAPTIVE AND NON-ADAPTIVE VERSIONS--- 
        The value of the scale parameter in the Laplace distribution.
        A common choice for the laplace prior is scale = 2/lambda, 
        because it can be shown that this is the MAP estimate where
        a laplace prior is equivalent to performing lasso regression.
        lambda is the L1 penalty, or scale = 2*C (where C is the penalty term
        in sklearn).
        
        I find that when scale == C, you get more similar results to
        the model output in sklearn, and based on
        my research into their implementation, I think it is because the
        sklearn implementation actually scales lambda by 1/2 (see the user
        guide here for more details:
        https://scikit-learn.org/stable/modules/linear_model.html),
        but i'm not 100% sure on this. This parameterization is similar to
        scale = stddev/lambda or scale = stddev*C, where they set stddev to
        1 instead of 2. The default value is 1.
    neg : bool, optional
        Return negative density. The default value is False.
    log : bool, optional
        Return log-density. The default value is True.

    Returns
    -------
    float
        The posterior density (height of the posterior density)
    '''
    n_mu = 0
    n_sigma = 1
    l_loc = 0
    # assuming an exponential prior for the scale parameter of the laplace
    # mean/scale of 0.38 seemed to work well for these experiments
    l_scale_rate = 1/0.38
    
    intercept = parms[0]
    betas = parms[1:-1]
    l_scale = np.exp(parms[-1])
    
    mu = X.dot(parms[:-1]).reshape(-1, )
    Ypred = 1/(1 + np.exp(-mu))
    like = np.sum(Y*np.log(Ypred) + (1 - Y)*np.log(1 - Ypred))
    # normal prior on the intercept
    int_prior = (-0.5*np.log(2.0*np.pi) - np.log(n_sigma) - (((intercept - n_mu)**2) / (2*(n_sigma**2))))
    # Laplace prior for Coefficient
    parm_prior = np.sum(-np.log(2*l_scale) - (np.abs(betas - l_loc)/l_scale))
    # exponential prior for scale parameter
    scale_prior = np.log(l_scale_rate) - l_scale_rate * l_scale
    # post = likelihood + laplace prior + jacobian for laplace scale + laplace scale prior
    post = like + int_prior + parm_prior + parms[-1] + scale_prior
    
    if not log:
        post = np.exp(post)

    if neg:
        return -post
    else:
        return post


def ap_poisson_lasso(parms, X, Y, l_scale=None, neg=False, log=True):
    '''
    The posterior density for a poisson regression model with an L1 penalty
    term. However, unlike the poisson_regression function, this is made to
    addaptively learn the optimal L1 penalty. Therefore, the L1 penalty is
    in the parms variable, at the end of the array
    
    Parameters
    ----------
    parms : numpy array (numeric)
        The model coefficients (including intercept, which is first, and the
        scale of the laplace distribution, which is last)
    X : numpy array (numeric)
        The independent variables (or feature matrix), where the first column
        is a dummy column of 1's (for the intercept).
    Y : numpy array or pandas dataframe
        The response value (should be 0 or 1, but could be float as well if 
        you're willing to deal with those consequences).
    l_scale : float, optional
        ---THIS IS NOT USED. ONLY HERE FOR CONVENIENCE OF THE USER WHEN
        EXPERIMENTING BETWEEN THE ADAPTIVE AND NON-ADAPTIVE VERSIONS--- 
        The value of the scale parameter in the Laplace distribution.
        A common choice for the laplace prior is scale = 2/lambda, 
        because it can be shown that this is the MAP estimate where
        a laplace prior is equivalent to performing lasso regression.
        lambda is the L1 penalty, or scale = 2*C (where C is the penalty term
        in sklearn).
        
        I find that when scale == C, you get more similar results to
        the model output in sklearn, and based on
        my research into their implementation, I think it is because the
        sklearn implementation actually scales lambda by 1/2 (see the user
        guide here for more details:
        https://scikit-learn.org/stable/modules/linear_model.html),
        but i'm not 100% sure on this. This parameterization is similar to
        scale = stddev/lambda or scale = stddev*C, where they set stddev to
        1 instead of 2. The default value is 1.
    neg : bool, optional
        Return negative density. The default value is False.
    log : bool, optional
        Return log-density. The default value is True.

    Returns
    -------
    float
        The posterior density (height of the posterior density)
    '''
    n_sigma = 10
    l_loc = 0
    # assuming an exponential prior for the scale parameter of the laplace
    # mean/scale of 0.38 seemed to work well for these experiments
    l_scale_rate = 1/0.38
    
    intercept = parms[0]
    betas = parms[1:-1]
    l_scale = np.exp(parms[-1])
    
    mu = X.dot(parms[:-1])
    
    like = np.sum(Y*mu - np.exp(mu) - loggamma(Y + 1))
    # normal prior on the intercept
    int_prior = (-0.5*np.log(2.0*np.pi) - np.log(n_sigma) - (((intercept - l_loc)**2) / (2.0*(n_sigma**2))))
    # Laplace prior 
    #parm_prior = np.sum(-np.log(2*l_scale) - (np.abs(betas - l_loc)/l_scale))
    # Normal prior
    parm_prior = np.sum(-0.5*np.log(2.0*np.pi) - np.log(l_scale) - (((betas - l_loc)**2) / (2.0*(l_scale**2))))
    
    # exponential prior for scale parameter
    scale_prior = np.log(l_scale_rate) - l_scale_rate * l_scale
    # post = likelihood + laplace prior + jacobian for laplace scale + laplace scale prior
    
    post = like + parm_prior + int_prior + parms[-1] + scale_prior
    
    if not log:
        post = np.exp(post)

    if neg:
        return -post
    else:
        return post


def ap_poisson_lasso_od(parms, X, Y, l_scale=None, neg=False, log=True):
    '''
    The posterior density for a poisson regression model with an L1 penalty
    term. However, unlike the poisson_regression function, this is made to
    addaptively learn the optimal L1 penalty. Therefore, the L1 penalty is
    in the parms variable, at the end of the array. This function also attempts
    to adaptively account for overdispersion.
    
    Parameters
    ----------
    parms : numpy array (numeric)
        The model coefficients (including intercept, which is first, and the
        scale of the laplace distribution, which is last)
    X : numpy array (numeric)
        The independent variables (or feature matrix), where the first column
        is a dummy column of 1's (for the intercept).
    Y : numpy array or pandas dataframe
        The response value (should be 0 or 1, but could be float as well if 
        you're willing to deal with those consequences).
    l_scale : float, optional
        ---THIS IS NOT USED. ONLY HERE FOR CONVENIENCE OF THE USER WHEN
        EXPERIMENTING BETWEEN THE ADAPTIVE AND NON-ADAPTIVE VERSIONS--- 
        The value of the scale parameter in the Laplace distribution.
        A common choice for the laplace prior is scale = 2/lambda, 
        because it can be shown that this is the MAP estimate where
        a laplace prior is equivalent to performing lasso regression.
        lambda is the L1 penalty, or scale = 2*C (where C is the penalty term
        in sklearn).
        
        I find that when scale == C, you get more similar results to
        the model output in sklearn, and based on
        my research into their implementation, I think it is because the
        sklearn implementation actually scales lambda by 1/2 (see the user
        guide here for more details:
        https://scikit-learn.org/stable/modules/linear_model.html),
        but i'm not 100% sure on this. This parameterization is similar to
        scale = stddev/lambda or scale = stddev*C, where they set stddev to
        1 instead of 2. The default value is 1.
    neg : bool, optional
        Return negative density. The default value is False.
    log : bool, optional
        Return log-density. The default value is True.

    Returns
    -------
    float
        The posterior density (height of the posterior density)
    '''
    n_sigma = 10
    # I think in pretty much every case you want this zero, so l_loc applies
    # to both the normal mu parameter and the laplace location parameter
    l_loc = 0
    # assuming an exponential prior for the scale parameter of the laplace
    # mean/scale of 0.38 seemed to work well for these experiments
    l_scale_rate = 1/0.38
    
    intercept = parms[0]
    betas = parms[1:-2]
    l_scale = np.exp(parms[-2])
    delta = parms[-1]
    
    mu = X.dot(parms[:-2]) - parms[-1]
    lamda = np.exp(mu)
    
    like = np.sum(mu + (Y - 1)*np.log(lamda + delta*Y) - lamda - delta*Y - loggamma(Y + 1))

    # normal prior on the intercept
    int_prior = (-0.5*np.log(2.0*np.pi) - np.log(n_sigma) - (((intercept - l_loc)**2) / (2.0*(n_sigma**2))))
    # Laplace prior 
    #parm_prior = np.sum(-np.log(2*l_scale) - (np.abs(betas - l_loc)/l_scale))
    # Normal prior
    parm_prior = np.sum(-0.5*np.log(2.0*np.pi) - np.log(l_scale) - (((betas - l_loc)**2) / (2.0*(l_scale**2))))
    
    # exponential prior for scale parameter
    scale_prior = np.log(l_scale_rate) - l_scale_rate * l_scale
    # post = likelihood + laplace prior + jacobian for laplace scale + laplace scale prior
    
    post = like + parm_prior + int_prior + parms[-2] + scale_prior
    
    if not log:
        post = np.exp(post)

    if neg:
        return -post
    else:
        return post


def posterior_poisson_lasso(parms, X, Y, l_scale=1, neg=False, log=True):
    '''
    The posterior density for a poisson regression model with an L1 penalty
    term.
    
    Parameters
    ----------
    parms : numpy array (numeric)
        The model coefficients (including intercept, which is first, and the
        scale of the laplace distribution, which is last)
    X : numpy array (numeric)
        The independent variables (or feature matrix), where the first column
        is a dummy column of 1's (for the intercept).
    Y : numpy array or pandas dataframe
        The response value (should be 0 or 1, but could be float as well if 
        you're willing to deal with those consequences).
    l_scale : float, optional
        The value of the scale parameter in the Laplace distribution.
        A common choice for the laplace prior is scale = 2/lambda, 
        because it can be shown that this is the MAP estimate where
        a laplace prior is equivalent to performing lasso regression.
        lambda is the L1 penalty, or scale = 2*C (where C is the penalty term
        in sklearn).
        
        I find that when scale == C, you get more similar results to
        the model output in sklearn, and based on
        my research into their implementation, I think it is because the
        sklearn implementation actually scales lambda by 1/2 (see the user
        guide here for more details:
        https://scikit-learn.org/stable/modules/linear_model.html),
        but i'm not 100% sure on this. This parameterization is similar to
        scale = stddev/lambda or scale = stddev*C, where they set stddev to
        1 instead of 2. The default value is 1.
    neg : bool, optional
        Return negative density. The default value is False.
    log : bool, optional
        Return log-density. The default value is True.

    Returns
    -------
    float
        The posterior density (height of the posterior density)
    '''
    n_sigma = 10
    # I think in pretty much every case you want this zero, so l_loc applies
    # to both the normal mu parameter and the laplace location parameter
    l_loc = 0
    
    intercept = parms[0]
    betas = parms[1:]
    
    mu = X.dot(parms)
    
    like = np.sum(Y*mu - np.exp(mu) - loggamma(Y + 1))
    # normal prior on the intercept
    int_prior = (-0.5*np.log(2.0*np.pi) - np.log(n_sigma) - (((intercept - l_loc)**2) / (2.0*(n_sigma**2))))
    # Laplace prior 
    #parm_prior = np.sum(-np.log(2*l_scale) - (np.abs(betas - l_loc)/l_scale))
    # Normal prior
    coef_prior = np.sum(-0.5*np.log(2.0*np.pi) - np.log(l_scale) - (((betas - l_loc)**2) / (2.0*(l_scale**2))))
    
    post = like + coef_prior + int_prior 
    
    if not log:
        post = np.exp(post)

    if neg:
        return -post
    else:
        return post


def posterior_poisson_lasso_od(parms, X, Y, l_scale=1, neg=False, log=True):
    '''
    The posterior density for a poisson regression model with an L1 penalty
    term. This function also attempts to adaptively account for overdispersion.
    
    Uses a generalized poisson model for the dispersion calculation. See
    https://journals.sagepub.com/doi/pdf/10.1177/1536867X1201200412
    
    Parameters
    ----------
    parms : numpy array (numeric)
        The model coefficients (including intercept, which is first, and the
        scale of the laplace distribution, which is last)
    X : numpy array (numeric)
        The independent variables (or feature matrix), where the first column
        is a dummy column of 1's (for the intercept).
    Y : numpy array or pandas dataframe
        The response value (should be 0 or 1, but could be float as well if 
        you're willing to deal with those consequences).
    l_scale : float, optional
        The value of the scale parameter in the Laplace distribution.
        A common choice for the laplace prior is scale = 2/lambda, 
        because it can be shown that this is the MAP estimate where
        a laplace prior is equivalent to performing lasso regression.
        lambda is the L1 penalty, or scale = 2*C (where C is the penalty term
        in sklearn).
        
        I find that when scale == C, you get more similar results to
        the model output in sklearn, and based on
        my research into their implementation, I think it is because the
        sklearn implementation actually scales lambda by 1/2 (see the user
        guide here for more details:
        https://scikit-learn.org/stable/modules/linear_model.html),
        but i'm not 100% sure on this. This parameterization is similar to
        scale = stddev/lambda or scale = stddev*C, where they set stddev to
        1 instead of 2. The default value is 1.
    neg : bool, optional
        Return negative density. The default value is False.
    log : bool, optional
        Return log-density. The default value is True.

    Returns
    -------
    float
        The posterior density (height of the posterior density)
    '''
    n_sigma = 10
    # I think in pretty much every case you want this zero, so l_loc applies
    # to both the normal mu parameter and the laplace location parameter
    l_loc = 0
    
    intercept = parms[0]
    betas = parms[1:-1]   
    
    # alternative formulation:
    # see: https://journals.sagepub.com/doi/pdf/10.1177/1536867X1201200412
    
    # custom bound on delta so it is strictly less than 1
    #delta = np.log(1/(1+np.exp(-parms[-1]))) + 1
    # else:
    delta = parms[-1]
    
    lamda = np.exp(X.dot(parms[:-1]))
    theta = lamda*(1 - delta)
    
    log_theta = np.log(theta)
    like = np.sum(log_theta + xlogy(Y - 1, theta + delta*Y) - theta - delta*Y - loggamma(Y + 1))
       
    # normal prior on the intercept
    int_prior = (-0.5*np.log(2.0*np.pi) - np.log(n_sigma) - (((intercept - l_loc)**2) / (2.0*(n_sigma**2))))
    # Laplace prior on beta
    coef_prior = np.sum(-np.log(2*l_scale) - (np.abs(betas - l_loc)/l_scale))
    # Normal prior on beta
    #coef_prior = np.sum(-0.5*np.log(2.0*np.pi) - np.log(l_scale) - (((betas - l_loc)**2) / (2.0*(l_scale**2))))
    
    ## prior on delta
    # normal
    delta_prior = (-0.5*np.log(2.0*np.pi) - np.log(0.4) - (((np.exp(delta) - 0)**2) / (2.0*(0.4**2))))
    
    # jacobian on custom delta transform
    # delta_jac = -np.log(np.exp(parms[-1]) + 1)

    post = (like
            + coef_prior
            + int_prior
            + delta_prior
            #+ delta_jac
            )
    
    if not log:
        post = np.exp(post)

    if neg:
        return -post
    else:
        return post


def weibull_regression_post(parms, X, Y, status=None, l_scale=1, neg=False,
                            log=True):
    '''
    The posterior density for a weibull regression model with an L1 penalty
    term. 
    
    Parameters
    ----------
    parms : numpy array (numeric)
        The model coefficients (including intercept, which is first, and the
        scale of the laplace distribution, which is last)
    X : numpy array (numeric)
        The independent variables (or feature matrix), where the first column
        is a dummy column of 1's (for the intercept).
    Y : numpy array or pandas dataframe
        The response value (should be 0 or 1, but could be float as well if 
        you're willing to deal with those consequences).
    status : numpy array (int, bool) or None
        Indicates status used for censoring, in this case, right censoring.
        An array of 1's and 0's, such that a 1 indicates an event, and 0 is the
        censored observation. If None, it will assume there are no censored
        events. Default is None.
    l_scale : float, optional
        The value of the scale parameter in the Laplace distribution.
        
        A common choice for the laplace prior is scale = 2/lambda, or
        scale = 2*C, where lambda is the L1 penalty (regularization strength)
        and C is the inverse penalty/strength. This is because it can be shown
        that this is the MAP estimate where a laplace prior is equivalent
        to performing lasso regression.
        
        I find that when scale == C, you get more similar results to
        the model output in sklearn, and based on
        my research into their implementation, I think it is because the
        sklearn implementation actually scales lambda by 1/2 (see the user
        guide here for more details:
        https://scikit-learn.org/stable/modules/linear_model.html),
        but i'm not 100% sure on this, and it deserves more exploration.
        This parameterization is similar to scale = variance/lambda or
        scale = variance*C, where they set variance to 1 instead of 2.
        The default value is 1.
    neg : bool, optional
        Return negative density. The default value is False.
    log : bool, optional
        Return log-density. The default value is True.

    Returns
    -------
    float
        The posterior density (height of the posterior density)
    '''
    l_loc = 0
    n_sigma = 10
    
    intercept = parms[0]
    betas = parms[1:-1]
    
    scale = np.exp(parms[-1])
    logshape = X.dot(parms[:-1])
    shape = np.exp(logshape)
    
    if status is None:
        status = np.repeat(1, X.shape[0])

    fails = Y[status == 1]
    cens = Y[status == 0]
    shape_fail = shape[status == 1]
    shape_cens = shape[status == 0]
    gshape = 0.01
    gscale = 100.0
    
    xly_1 = xlogy(shape - 1.0, fails/scale)
    xly_2 = xlogy(gshape - 1.0, scale)
    
    if fails.shape[0] != 0:
        fail_like = np.sum((np.log(shape_fail) - np.log(scale) + xly_1 - (fails/scale)**shape))
    else:
        fail_like = 0
        
    if cens.shape[0] != 0:
        cens_like = np.sum(-(cens/scale)**shape_cens)
    else:
        cens_like = 0
    
    #scale prior
    scale_prior = (-(loggamma(gshape) + gshape*np.log(gscale)) + xly_2 - (scale/gscale))
    # normal prior on the intercept
    int_prior = (-0.5*np.log(2.0*np.pi) - np.log(n_sigma) - (((intercept - l_loc)**2) / (2.0*(n_sigma**2))))
    # Laplace prior 
    parm_prior = np.sum(-np.log(2*l_scale) - (np.abs(betas - l_loc)/l_scale))
    # Normal prior
    #parm_prior = np.sum(-0.5*np.log(2.0*np.pi) - np.log(l_scale) - (((betas - l_loc)**2) / (2.0*(l_scale**2))))
    # assuming uniform for over dispersion prior
    post = fail_like + cens_like + parm_prior + int_prior + scale_prior + parms[-1]
    
    if not log:
        post = np.exp(post)

    if neg:
        return -post
    else:
        return post


def pois_uniform(param, count, neg=False, log=True):
    '''
    Posterior density for the Poisson distribution (assuming uniform prior)

    Parameters
    ----------
    param : float
        log of the Poisson mean (lambda).
    count : int, or array of ints
        count of events (or poisson counts)
    neg : bool, optional
        Return negative density. The default value is False.
    log : bool, optional
        Return log-density. The default value is True.

    Returns
    -------
    float
        The posterior density (height of the posterior density).

    '''
    lmbda = np.exp(param)
    like = np.sum(count * param - lmbda - loggamma(count + 1.0))
    jac = param
    
    post = like + jac
    
    if not log:
        post = np.exp(post)

    if neg:
        return -post
    else:
        return post


def pois_gamma(param, count, gshape=0.01, gscale=100, neg=False, log=True):
    '''
    Posterior density for the Poisson distribution (assuming gamma prior)

    Parameters
    ----------
    param : float
        Log of the Poisson mean (lambda).
    count : int, or array of ints
        count of events (or poisson counts)
    gshape : float
        shape parameter of the gamma prior
    gscale : float
        scale parameter of the gamma prior
    neg : bool, optional
        Return negative density. The default value is False.
    log : bool, optional
        Return log-density. The default value is True.

    Returns
    -------
    float
        The posterior density (height of the posterior density).

    '''
    lmbda = np.exp(param)
    like = np.sum(count * param - lmbda - loggamma(count + 1.0))
    prior = (-(loggamma(gshape) + gshape*np.log(gscale)) +
             xlogy(gshape-1, lmbda) - (lmbda/gscale))
    jac = param
    
    post = like + prior + jac
    
    if not log:
        post = np.exp(post)

    if neg:
        return -post
    else:
        return post


def pois_gamma_ada(param, count, neg=False, log=True):
    '''
    Posterior density for the Poisson distribution (assuming gamma prior),
    learns the gamma shape/scale as part of mcmc

    Parameters
    ----------
    param : numpy array (float)
        The log of the Poisson mean (lambda), and the shape/scale
        of the gamma prior.
    count : int, or array of ints
        count of events (or poisson counts)
    neg : bool, optional
        Return negative density. The default value is False.
    log : bool, optional
        Return log-density. The default value is True.

    Returns
    -------
    float
        The posterior density (height of the posterior density).

    '''
    lmbda = np.exp(param[0])
    gshape = np.exp(param[1])
    gscale = np.exp(param[2])
    like = np.sum(count * param[0] - lmbda - loggamma(count + 1.0))
    prior = (-(loggamma(gshape) + gshape*np.log(gscale)) +
             xlogy(gshape-1, lmbda) - (lmbda/gscale))
    jac = np.sum(param)
    
    post = like + prior + jac
    
    if not log:
        post = np.exp(post)

    if neg:
        return -post
    else:
        return post


def exp_gamma(param, data, status, gshape=0.01, gscale=100, neg=False,
              log=True):
    '''
    Posterior distribution for an exponential likelihood with a gamma prior.
    Allows for censoring.

    Parameters
    ----------
    param : float
        The log of the rate parameter of an exponential distribution.
    data : numpy array
        The data for the exponential likelihood (Usually times).
    status : numpy array
        An array of binary variables indicating which items are censored (1's are events, 0's are unknown). If there
        are no censored observations, pass an array of 1's.
    gshape : float
        shape parameter of the gamma prior
    gscale : float
        scale parameter of the gamma prior
    neg : bool, optional
        Return negative density. The default value is False.
    log : bool, optional
        Return log-density. The default value is True.

    Returns
    -------
    post : float
        The posterior density.

    '''
    if status is None:
        status = np.repeat(1, data.shape[0])
    
    rate = np.exp(param)
    events = data[status == 1]
    cens = data[status == 0]
    post = ((len(events) * param) -
            (rate*(np.sum(events) +
                   np.sum(cens))) +
            (-(loggamma(gshape) + gshape*np.log(gscale)) +
             ((gshape - 1.0)*param) - (rate/gscale)) +
            param)

    if not log:
        post = np.exp(post)

    if neg:
        return -post
    else:
        return post


def weibull_gamma(param, data, status=None, gshape=0.01, gscale=100,
                  neg=False, log=True):
    '''
    The posterior density for a weibull distribution with a gamma prior

    Parameters
    ----------
    param : numpy array (float)
        The log shape and scale parameters
    data : numpy array (float)
        The data being fit
    status : numpy array (int, or bool), optional
        Indicates status used for censoring, in this case, right censoring.
        An array of 1's and 0's, such that a 1 indicates an event, and 0 is the
        censored observation. If None, it will assume there are no censored
        events. The default is None.
    gshape : float
        shape parameter of the gamma prior
    gscale : float
        scale parameter of the gamma prior
    neg : bool, optional
        Return negative density. The default value is False.
    log : bool, optional
        Return log-density. The default value is True.

    Returns
    -------
    float
        The posterior density (height of the posterior density).

    '''
    shape = np.exp(param[0])
    scale = np.exp(param[1])

    if status is None:
        status = np.repeat(1, data.shape[0])

    fails = data[status == 1]
    cens = data[status == 0]

    xly_1 = (shape - 1.0)*np.log(fails/scale)
    xly_2 = (gshape - 1.0)*np.log(scale)
    if (shape - 1.0) == 0:
        xly_1[np.where(np.isnan(xly_1))] = 0

    fail_like = (np.log(shape) - np.log(scale) +
                 xly_1 - (fails/scale)**shape)
    cens_like = -(cens/scale)**shape

    prior = (-(loggamma(gshape) + gshape*np.log(gscale)) +
             xly_2 - (scale/gscale))

    post = np.sum(fail_like) + np.sum(cens_like) + prior + param[0] + param[1]
    
    if not log:
        post = np.exp(post)

    if neg:
        return -post
    else:
        return post


def NHPP_posterior(lparm, tints, tbar, obs, a, b, mu, sigma, neg=False,
                   log=True):
    '''
    The posterior density of the non-homogenous Poisson distribution with
    a power-law process. 

    Parameters
    ----------
    lparm : numpy array (float)
        log parameters eta and phi (shape and scale with a variable change).
    tints : numpy array (int)
        The sequence of time intervals (1,2,3, etc...) for the poisson counts.
    tbar : int
        time between tints (usually 1).
    obs : numpy array (int)
        The data (usually count) for the poisson process (equal to length as tints).
    a : float
        lower bound of the uniform prior on phi.
    b : float
        upper bound of the uniform prior on phi.
    mu : float
        This is a paramter that is part of the variable change on the gamma
        prior for the poisson likelihood.
    sigma : float
        This is a paramter that is part of the variable change on the gamma
        prior for the poisson likelihood.
    neg : bool, optional
        Return negative density. The default value is False.
    log : bool, optional
        Return log-density. The default value is True.

    Returns
    -------
    float
        The posterior density (height of the posterior density).

    '''
    # The Log postirior for NHPP with rate = (phi/eta)*(t/eta)**(phi-1) using
    # only Math no premade distributions.

    # Preform variable change.
    shape = (np.exp(lparm[0]) * b + a) / (1 + np.exp(lparm[0]))
    scale = np.exp(lparm[1])

    # Extrace interval start and end times from time series.
    int_start = tints[0:len(tints)-1]
    int_end = tints[1:len(tints)]

    # Clculate log prior
    # This prior is a Gamma distribution with shape = (mu/sigma)^2,
    # Rate = (mu/sigma^2), a verible change of x = (t_bar/eta)^phi was made
    # with variable eta and phi and t_bar as constants. So the end prior is
    # Gamma((t_bar/eta)^phi,(mu/sigma)^2,(mu/sigma^2))*(dx/deta)*Uniform(phi,a,b)
    lp = (np.log(shape) + (mu/sigma)*(mu/sigma)*np.log(mu/(sigma*sigma)) +
          (shape*(mu/sigma)*(mu/sigma))*(np.log(tbar)-np.log(scale)) -
          np.log(scale) - (mu/(sigma*sigma))*((tbar/scale)**shape) -
          loggamma((mu/sigma)*(mu/sigma)))
    lp_unif = 0.0
    if a >= b:
        lp_unif = np.nan
    elif a <= shape <= b:
        lp_unif = -np.log(b - a)

    lprior = lp + lp_unif

    # Calulate log Liklihood
    llik = np.sum(
            obs*(np.log(((int_end/scale)**shape) -
                        ((int_start/scale)**shape))) -
                (((int_end/scale)**shape) -
                 ((int_start/scale)**shape)) - loggamma(obs + 1))

    # Calulate Log Jacobian from the above variable change.
    # the term np.log(b-a) was removed because it's just a constant.
    ljac = (lparm[0] + lparm[1] - 2.0*np.log(1.0 + np.exp(lparm[0])))
    
    post = llik + lprior + ljac
    
    if not log:
        post = np.exp(post)
    # Sum and return
    if neg:
        return -post
    else:
        return post

    